import re
import math
from decimal import Decimal

import numpy as np
import pandas as pd
from scipy.stats import distributions

import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from itertools import groupby

# boxplot
def _get_boxplot_quantiles(series, whis=1.5):
    # get the matplotlib boxplot quantiles
    # sort of direct copy from the matplotlib library

    q1, median, q3 = np.percentile(series, [25, 50, 75])

    if np.isscalar(whis):
        iqr = q3 - q1
        loval = q1 - (whis * iqr)
        hival = q3 + (whis * iqr)
    else:
        loval, hival = np.percentile(series, sorted(whis))

    # get high extreme
    wiskhi = series[series <= hival]
    if len(wiskhi) == 0 or np.max(wiskhi) < q3:
        high = q3
    else:
        high = np.max(wiskhi)

    # get low extreme
    wisklo = series[series >= loval]
    if len(wisklo) == 0 or np.min(wisklo) > q1:
        low = q1
    else:
        low = np.min(wisklo)
    
    return [low, q1, q3, high]

def _get_boxplot_outliers(series, whis=1.5):
    low, q1, q3, high = _get_boxplot_quantiles(series=series, whis=whis)

    return series[series < low].append(series[series > high]).values


# histogram
def _bins(series, points, min_bins=10, max_bins=20, max_decimals=10):

    if not isinstance(series, pd.Series):
        raise ValueError('series is not a Pandas Series')

    if not isinstance(points, (tuple, list)):
        raise ValueError('points is not a tuple or a list')

    if not isinstance(min_bins, int):
        raise ValueError('min_bins is not an integer')
    else:
        if min_bins < 1:
            min_bins = 1
    
    if not isinstance(max_bins, int):
        raise ValueError('max_bins is not an integer')
    else:
        if max_bins < 5:
            max_bins = 5
    
    
    # decimal places for points
    decimal_places = min(-min([Decimal(str(x)).as_tuple().exponent for x in points]), max_decimals)
    
    gcd_factor = 10**decimal_places
    bin_size = np.gcd.reduce([int(round(x*gcd_factor)) for x in points]) / gcd_factor
    
    # adjust bin_size
    series_range = series.max() - series.min()

    ## min_bins
    range_factor = np.ceil(min_bins/(series_range/bin_size))
    bin_size = bin_size / range_factor
    
    ## max_bins
    range_factor = np.ceil((series_range/bin_size)/max_bins)
    bin_size = bin_size * range_factor

    # new decimal places
    decimal_places = -Decimal(str(bin_size)).as_tuple().exponent
    
    # get start and end
    start = round(series.min() - (series.min() % bin_size), decimal_places)
    end = round(series.max() - (series.max() % bin_size) + bin_size, decimal_places)

    return np.arange(start, end, bin_size)



# matplotlib format
def _axisformat(ax, series):

    if isinstance(series.index, pd.DatetimeIndex):
        rng = series.index.max() - series.index.min()

        if rng > pd.Timedelta('365 days'):
            ax.xaxis.set_major_locator(mdates.MonthLocator(interval=3))
            ax.xaxis.set_major_formatter(mdates.DateFormatter('%m-%Y'))
            ax.set_xlabel('month')
        elif rng > pd.Timedelta('183 days'):
            ax.xaxis.set_major_locator(mdates.MonthLocator())
            ax.xaxis.set_major_formatter(mdates.DateFormatter('%b'))
            ax.set_xlabel('month')
        elif rng > pd.Timedelta('31 days'):
            ax.xaxis.set_major_locator(mdates.WeekdayLocator())
            ax.xaxis.set_major_formatter(mdates.DateFormatter('%V'))
            ax.set_xlabel('week')
        else:
            ax.xaxis.set_major_locator(mdates.DayLocator(interval=7))
            ax.xaxis.set_major_formatter(mdates.DateFormatter('%d-%m-%Y'))
            ax.set_xlabel('date')
    elif isinstance(series, pd.DataFrame):
        ax.set_xlabel(series.iloc[:,0].name)
        ax.set_ylabel(series.index.name)
    else:
        ax.set_xlabel(series.name)

    return ax

def _get_multiindex_labels(ax, _obj, style):
    # great thanks to Trenton McKinney
    # https://stackoverflow.com/questions/19184484/how-to-add-group-labels-for-bar-charts-in-matplotlib

    def _add_line(ax, xpos, ypos):
        line = plt.Line2D([xpos, xpos], [ypos + .1, ypos], transform=ax.transAxes, **style.graphs.line.multiindex_label_line)
        line.set_clip_on(False)
        ax.add_line(line)

    def _get_label_len(my_index,level):
        labels = my_index.get_level_values(level)
        return [(k, sum(1 for i in g)) for k,g in groupby(labels)]

    # remove current labels
    ax.set_xticklabels([])
    ax.set_xlabel('')

    # draw new labels
    ypos = -.1
    scale = 1./_obj.index.size
    for level in range(_obj.index.nlevels)[::-1]:
        pos = 0
        for label, rpos in _get_label_len(_obj.index,level):
            lxpos = (pos + .5 * rpos)*scale
            ax.text(lxpos, ypos, label, ha='center', transform=ax.transAxes)
            _add_line(ax, pos*scale, ypos)
            pos += rpos
        _add_line(ax, pos*scale , ypos)
        ypos -= .1

def _get_nice_list(the_list, max_len=3):
    if len(the_list) == 1:
        return the_list[0]
    elif len(the_list) <= max_len:
        return ", ".join(the_list[:-1])+" and "+the_list[-1]
    else:
        return ", ".join(the_list[:2])+", ... and "+the_list[-1]

def _get_valid_filename(s):
    s = str(s).strip().replace(' ', '_')
    return re.sub(r'(?u)[^-\w.]', '', s)

def _get_digits(x, digits=6):
    if x > 0:
        prtd = max(digits, math.ceil(math.log10(abs(x))))
        prtd -= math.ceil(math.log10(abs(x)))
        prtd = min(digits, prtd)
        return prtd
    else:
        return 0
        
def _format_digits(x, digits=6):
    if isinstance(x, list):
        return _format_digits(max([_get_digits(i, digits=digits) for i in x if i is not None]), digits=digits)

    elif x is None:
        return 0
    elif x == 0 or not math.isfinite(x):
        return x
    else:
        prtd = max(digits, math.ceil(math.log10(abs(x))))
        prtd -= math.ceil(math.log10(abs(x)))
        prtd = min(digits, prtd)
        return f'1.{prtd}f'

def _get_h0_equal_means(columns):
    if len(columns) == 1:
        return ''
    elif len(columns) > 4:
        return f'$H_0: \\bar{{X}}_{{{columns[0]}}} = \\bar{{X}}_{{{columns[1]}}} = \\bar{{X}}_{{...}} = \\bar{{X}}_{{{columns[-1]}}}$'
    else:
        result = f'$H_0: \\bar{{X}}_{{{columns[0]}}}$'
        for col in columns[1:]:
            result += f'$ = \\bar{{X}}_{{{col}}}$'
        return result
    
def _get_distribution(dist=None):
    distributions = {
        "norm": "Normal (Gaussian)",
        "alpha": "Alpha",
        "anglit": "Anglit",
        "arcsine": "Arcsine",
        "beta": "Beta",
        "betaprime": "Beta Prime",
        "bradford": "Bradford",
        "burr": "Burr",
        "cauchy": "Cauchy",
        "chi": "Chi",
        "chi2": "Chi-squared",
        "cosine": "Cosine",
        "dgamma": "Double Gamma",
        "dweibull": "Double Weibull",
        "erlang": "Erlang",
        "expon": "Exponential",
        "exponweib": "Exponentiated Weibull",
        "exponpow": "Exponential Power",
        "fatiguelife": "Fatigue Life (Birnbaum-Sanders)",
        "foldcauchy": "Folded Cauchy",
        "f": "F (Snecdor F)",
        "fisk": "Fisk",
        "foldnorm": "Folded Normal",
        "frechet_r": "Frechet Right Sided, Extreme Value Type II",
        "frechet_l": "Frechet Left Sided, Weibull_max",
        "gamma": "Gamma",
        "gausshyper": "Gauss Hypergeometric",
        "genexpon": "Generalized Exponential",
        "genextreme": "Generalized Extreme Value",
        "gengamma": "Generalized gamma",
        "genlogistic": "Generalized Logistic",
        "genpareto": "Generalized Pareto",
        "genhalflogist": "Generalized Half Logistic",
        "gilbrat": "Gilbrat",
        "gompertz": "Gompertz (Truncated Gumbel)",
        "gumbel_l": "Left Sided Gumbel, etc.",
        "gumbel_r": "Right Sided Gumbel",
        "halfcauchy": "Half Cauchy",
        "halflogistic": "Half Logistic",
        "halfnorm": "Half Normal",
        "hypsecant": "Hyperbolic Secant",
        "invgamma": "Inverse Gamma",
        "invnorm": "Inverse Normal",
        "invweibull": "Inverse Weibull",
        "johnsonsb": "Johnson SB",
        "johnsonsu": "Johnson SU",
        "laplace": "Laplace",
        "logistic": "Logistic",
        "loggamma": "Log-Gamma",
        "loglaplace": "Log-Laplace (Log Double Exponential",
        "lognorm": "Log-Normal",
        "lomax": "Lomax (Pareto of the second kind)",
        "maxwell": "Maxwell",
        "mielke": "Mielke's Beta-Kappa",
        "nakagami": "Nakagami",
        "ncx2": "Non-central chi-squared",
        "ncf": "Non-central F",
        "nct": "Non-central Student's T",
        "pareto": "Pareto",
        "powerlaw": "Power-function",
        "powerlognorm": "Power log normal",
        "powernorm": "Power normal",
        "rdist": "R distribution",
        "reciprocal": "Reciprocal",
        "rayleigh": "Rayleigh",
        "rice": "Rice",
        "recipinvgauss": "Reciprocal Inverse Gaussian",
        "semicircular": "Semicircular",
        "t": "Student's T",
        "triang": "Triangular",
        "truncexpon": "Truncated Exponential",
        "truncnorm": "Truncated Normal",
        "tukeylambda": "Tukey-Lambda",
        "uniform": "Uniform",
        "vonmises": "Von-Mises (Circular)",
        "wald": "Wald",
        "weibull_min": "Minimum Weibull (see Frechet)",
        "weibull_max": "Maximum Weibull (see Frechet)",
        "wrapcauchy": "Wrapped Cauchy",
        "ksone": "Kolmogorov-Smirnov one-sided (no st",
        "kstwobign": "Kolmogorov-Smirnov two-sided test for Large N",
    }

    return distributions.get(dist, None)


import bluebelt.styles.defaults as defaults

# main plot | pyplot.plot
plot = {
    'marker': None,
    'zorder': 20,
    'linestyle': '-',
    'linewidth': 1,
    'color': defaults.black,
}

# main plot | pyplot.axhline
axhline = {
    'marker': None,
    'zorder': 20,
    'linestyle': '--',
    'linewidth': 1,
    'color': defaults.blue,
}


optimum_fill_between = {
    'color': None,
    'edgecolor': defaults.blue,
    'facecolor': None,
    'hatch': '\\\\\\',
    'linestyle': 'dashed',
    'linewidth': 0.5,
}

# text | pyplot.text
text = {
    'backgroundcolor': defaults.white,
    'color': defaults.blue,
    'va': 'center',
    'ha': 'left',
    'size': defaults.small,
    'zorder': 99,
}

# bounds_text | pyplot.text
bounds_text = {
    'backgroundcolor': defaults.white,
    'color': defaults.blue,
    'va': 'top',
    'ha': 'left',
    'size': defaults.small,
    'zorder': 99,
}

# plot title | pyplot.set_title
title = {
    'loc': 'left',
    'size': defaults.medium
}



import bluebelt.styles.defaults as defaults


histogram = {
    'density': True, 
    'alpha': 1, # alpha fraction 0-1
    'edgecolor': None, # color name, hex or None
    'facecolor': defaults.blue, # color name, hex or None
    'fill': False, #True, False
    'hatch': '//', # '/', '\', '|', '-', '+', 'x', 'o', 'O', '.' or '*'
    'linestyle': 'solid', # '-', '--', '-.', ':', '', (offset, on-off-seq), ...
    'linewidth': 1, # float
}

# histogram patches that need to be split; in range specs | pyplot.hist
histogram_fill_in_range = {
    'hatch': '//', # '/', '\', '|', '-', '+', 'x', 'o', 'O', '.' or '*'
}

# histogram patches that need to be split; out of range specs | pyplot.hist
histogram_fill_out_of_range = {
    'edgecolor': defaults.red, # color name, hex or None
    'linewidth': 0,
    'hatch': '////', # '/', '\', '|', '-', '+', 'x', 'o', 'O', '.' or '*'
}

histogram_patch_edge_color = defaults.red
histogram_patch_edge_hatch = '////'

# plot used for normal distribution | pyplot.plot
normal_plot_overall = {
    'color': defaults.blue,
    'linestyle': 'dashed',
    'linewidth': 1,
}
normal_plot_within = {
    'color': defaults.red,
    'linestyle': 'dashed',
    'linewidth': 1,
}

# target axvline | pyplot.axvline
target_axvline = {
    'color': defaults.blue,
    'linestyle': 'dotted',
    'linewidth': 1,
}

# USL, LSL axvline | pyplot.axvline
sl_axvline = {
    'color': defaults.blue,
    'linestyle': 'dotted',
    'linewidth': 1,
}

# USL, LSL, target text | pyplot.text
sl_text = {
    'backgroundcolor': defaults.white,
    'va': 'center',
    'ha': 'center', 
}

# plot title | pyplot.set_title
title = {
    'loc': 'left',
}


import bluebelt.styles.defaults as defaults

# histogram | pyplot.hist
histogram = {
    'density': True, 
    'alpha': 1, # alpha fraction 0-1
    'edgecolor': None, # color name, hex or None
    'facecolor': defaults.blue, # color name, hex or None
    'fill': False, #True, False
    'hatch': '//', # '/', '\', '|', '-', '+', 'x', 'o', 'O', '.' or '*'
    'linestyle': 'solid', # '-', '--', '-.', ':', '', (offset, on-off-seq), ...
    'linewidth': 1, # float
}

# text | pyplot.text
text = {
    'backgroundcolor': defaults.white,
    'va': 'center',
    'ha': 'center',
    'zorder': 10,
}

# plot title | pyplot.set_title
title = {
    'loc': 'left',
    'size': defaults.small
}

# plot title | pyplot.set_title
suptitle = {
    'ha': 'left',
    'size': defaults.medium
}



import bluebelt.styles.defaults as defaults

# boxplot | pyplot.boxplot
boxplot = {
    'boxes': {
        'hatch': '//',
        },
}

# mean axvline | pyplot.axvline
mean_axvline = {
    'color': defaults.blue,
    'linestyle': 'dotted',
    'linewidth': 1,
}

# population mean axvline | pyplot.axvline
popmean_axvline = {
    'color': defaults.red,
    'linestyle': 'dotted',
    'linewidth': 1,
}

# ci median horizontal line | pyplot.plot
ci_median_plot = {
    'color': defaults.black,
}

# ci median vertical lines (whiskers) | pyplot.plot
ci_median_axvline = {
    'color': defaults.black,
}

# ci median scatter (mean point) : pyplot.scatter
ci_median_scatter = {
    'color': defaults.black,
}

# result text | pyplot.text
result_text = {
    'backgroundcolor': defaults.white,
    'va': 'top',
    'ha': 'left', 
}

# text | pyplot.text
text = {
    'backgroundcolor': defaults.white,
    'va': 'center',
    'ha': 'center',
    'zorder': 10,
}

# text | pyplot.text
text_ci_min = {
    'backgroundcolor': defaults.white,
    'va': 'center',
    'ha': 'right',
    'zorder': 0,
}

# text | pyplot.text
text_ci_max = {
    'backgroundcolor': defaults.white,
    'va': 'center',
    'ha': 'left',
    'zorder': 0, 
}

# plot title | pyplot.set_title
title = {
    'loc': 'left',
}


import pandas as pd
import numpy as np
import scipy.stats as stats


import bluebelt.datetime.dt

import bluebelt.statistics.std

import bluebelt.data.index
import bluebelt.data.subsets

import bluebelt.analysis.pattern
import bluebelt.analysis.forecast
import bluebelt.statistics.hypothesis_testing

import bluebelt.core.decorators

@bluebelt.core.decorators.class_methods
@pd.api.extensions.register_dataframe_accessor('blue')
class DataFrameToolkit():
    def __init__(self, pandas_obj):
        self._obj = pandas_obj
        self.data = self.data(self._obj)
        self.datetime = self.datetime(self._obj)
        self.forecast = self.forecast(self._obj)
        self.statistics = self.statistics(self._obj)
        self.pattern = self.pattern(self._obj)
        self.test = self.test(self._obj)
        self.graph = self.graph(self._obj)

    # first level functions

    def iso_index(self, **kwargs):
        return bluebelt.data.index.iso_index(self._obj, **kwargs)

    def dt_index(self, **kwargs):
        return bluebelt.data.index.dt_index(self._obj, **kwargs)

    def subset(self, inverse=False, **kwargs):
        return bluebelt.data.subsets.Subset(self._obj, inverse=inverse, **kwargs)

    @bluebelt.core.decorators.class_methods
    class data():
        def __init__(self, pandas_obj):
            self._obj = pandas_obj
    
        def subset(self, inverse=False, **kwargs):
            return bluebelt.data.subsets.Subset(self._obj, inverse=inverse, **kwargs)
        
        @bluebelt.core.decorators.add_docstring(bluebelt.data.resolution.GroupByDatetimeIndex)
        def group_index(self, **kwargs):
            return bluebelt.data.resolution.GroupByDatetimeIndex(self._obj, **kwargs)
        
        dt_index = group_index

    @bluebelt.core.decorators.class_methods
    class datetime():
        def __init__(self, pandas_obj):
            self._obj = pandas_obj
    
        # bluebelt.datetime.dt
        def year(self, column=None, **kwargs):
            return bluebelt.datetime.dt.year(self._obj[column], **kwargs)

        def quarter(self, column=None, **kwargs):
            return bluebelt.datetime.dt.quarter(self._obj[column], **kwargs)

        def month(self, column=None, **kwargs):
            return bluebelt.datetime.dt.month(self._obj[column], **kwargs)

        def day(self, column=None, **kwargs):
            return bluebelt.datetime.dt.day(self._obj[column], **kwargs)

        def weekday(self, column=None, **kwargs):
            return bluebelt.datetime.dt.weekday(self._obj[column], **kwargs)
        
        def weekday_name(self, column=None, **kwargs):
            return bluebelt.datetime.dt.weekday_name(self._obj[column], **kwargs)

        def is_holiday(self, column=None, **kwargs):
            return bluebelt.datetime.dt.is_holiday(self._obj[column], **kwargs)

        def is_weekend(self, column=None, **kwargs):
            return bluebelt.datetime.dt.is_weekend(self._obj[column], **kwargs)
        
        def iso_year(self, column=None, **kwargs):
            return bluebelt.datetime.dt.iso_year(self._obj[column], **kwargs)

        def iso_week(self, column=None, **kwargs):
            return bluebelt.datetime.dt.iso_week(self._obj[column], **kwargs)
        
        def week(self, column=None, **kwargs):
            return bluebelt.datetime.dt.week(self._obj[column], **kwargs)
        
        def date_from_weeknumber(self, year=None, week=None, **kwargs):
            return bluebelt.datetime.dt.date_from_weeknumber(self._obj, year=year, week=week, **kwargs)

        def add(self, column=None, prefix='_', **kwargs):
            self._obj.loc[:,f'{prefix}year'] = bluebelt.datetime.dt.year(self._obj[column])
            self._obj.loc[:,f'{prefix}quarter'] = bluebelt.datetime.dt.quarter(self._obj[column])
            self._obj.loc[:,f'{prefix}month'] = bluebelt.datetime.dt.month(self._obj[column])
            self._obj.loc[:,f'{prefix}day'] = bluebelt.datetime.dt.day(self._obj[column])
            self._obj.loc[:,f'{prefix}weekday'] = bluebelt.datetime.dt.weekday(self._obj[column])
            self._obj.loc[:,f'{prefix}day_name'] = bluebelt.datetime.dt.day_name(self._obj[column])
            self._obj.loc[:,f'{prefix}is_holiday'] = bluebelt.datetime.dt.is_holiday(self._obj[column])
            self._obj.loc[:,f'{prefix}is_weekend'] = bluebelt.datetime.dt.is_weekend(self._obj[column])
            self._obj.loc[:,f'{prefix}iso_year'] = bluebelt.datetime.dt.iso_year(self._obj[column])
            self._obj.loc[:,f'{prefix}iso_week'] = bluebelt.datetime.dt.iso_week(self._obj[column])
            return self._obj

    @bluebelt.core.decorators.class_methods
    class statistics():
        def __init__(self, pandas_obj):
            self._obj = pandas_obj
        
        # rolling std
        def rolling_std(self, column=None, window=7, center=True, **kwargs):
            if column in self._obj:
                return bluebelt.statistics.std.RollingStd(self._obj[column], window=window, center=center, **kwargs)
            elif column is not None:
                raise ValueError(f'{column} is not in the dataframe')
            else:
                raise ValueError(f'provide a column name')

        @bluebelt.core.decorators.add_docstring(bluebelt.statistics.std.StdWithin)
        def std_within(self, axis=0, columns=None, how=None, observations=2, **kwargs):
            # frame, axis=0, columns=None, how=None, observations=2, 
            return bluebelt.statistics.std.StdWithin(self._obj, axis=axis, columns=columns, how=how, observations=observations, **kwargs)

        def correlation(self, columns=None, confidence=0.95, **kwargs):
            return bluebelt.statistics.basic_statistics.Correlation(self._obj, columns=columns, confidence=confidence)

    @bluebelt.core.decorators.class_methods
    class pattern():
        def __init__(self, pandas_obj):
            self._obj = pandas_obj

        # patterns
        def anomalies(self, values=None, pattern=None):
            return bluebelt.analysis.pattern.anomalies(self._obj, values=values, pattern=pattern)

    @bluebelt.core.decorators.class_methods
    class forecast():

        def __init__(self, pandas_obj):
            self._obj = pandas_obj
                
        @bluebelt.core.decorators.add_docstring(bluebelt.analysis.forecast.MAPE)
        def MAPE(self, **kwargs):
            return bluebelt.analysis.forecast.MAPE(self._obj, **kwargs)

        @bluebelt.core.decorators.add_docstring(bluebelt.analysis.forecast.WMAPE)
        def WMAPE(self, **kwargs):
            return bluebelt.analysis.forecast.WMAPE(self._obj, **kwargs)

        @bluebelt.core.decorators.add_docstring(bluebelt.analysis.forecast.SMAPE)
        def SMAPE(self, **kwargs):
            return bluebelt.analysis.forecast.SMAPE(self._obj, **kwargs)
        
        @bluebelt.core.decorators.add_docstring(bluebelt.analysis.forecast.MDA)
        def MDA(self, **kwargs):
            return bluebelt.analysis.forecast.MDA(self._obj, **kwargs)

        @bluebelt.core.decorators.add_docstring(bluebelt.analysis.forecast.MPE)
        def MPE(self, **kwargs):
            return bluebelt.analysis.forecast.MPE(self._obj, **kwargs)

        mape = MAPE
        wmape = WMAPE
        wMAPE = WMAPE
        smape = SMAPE
        mda = MDA
        mpe = MPE
     
    @bluebelt.core.decorators.class_methods
    class test():
        def __init__(self, pandas_obj):
            self._obj = pandas_obj

        def anderson_darling(self, columns=None, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.AndersonDarling(self._obj, columns=columns, alpha=alpha, **kwargs)
        
        def normal_distribution(self, columns=None, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.NormalDistribution(self._obj, columns=columns, alpha=alpha, **kwargs)
        
        def dagostino_pearson(self, columns=None, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.DAgostinoPearson(self._obj, columns=columns, alpha=alpha, **kwargs)
        
        # index
        @property
        def index(self):
            return bluebelt.statistics.hypothesis_testing.index()

        # hypothesis testing
        
        def equal_means(self, columns=None, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.EqualMeans(self._obj, columns=columns, alpha=alpha, **kwargs)
        
        def two_sample_t(self, columns=None, related=False, confidence=0.95, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.TwoSampleT(self._obj, columns=columns, related=related, confidence=confidence, alpha=alpha, **kwargs)
        
        def wilcoxon(self, popmean=None, columns=None, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.Wilcoxon(self._obj, columns=columns, alpha=alpha, **kwargs)
        
        def anova(self, columns=None, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.Anova(self._obj, columns=columns, alpha=alpha, **kwargs)

        def kruskal(self, columns=None, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.KruskalWallis(self._obj, columns=columns, alpha=alpha, **kwargs)

        kruskal_wallis = kruskal

        def mann_whitney(self, columns=None, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.MannWhitney(self._obj, columns=columns, alpha=alpha, **kwargs)
        
        def levene(self, columns=None, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.Levene(self._obj, columns=columns, alpha=alpha, **kwargs)
        
        def bartlett(self, columns=None, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.Bartlett(self._obj, columns=columns, alpha=alpha, **kwargs)
        
        def f_test(self, columns=None, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.FTest(self._obj, columns=columns, alpha=alpha, **kwargs)
        
        def equal_variances(self, columns=None, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.EqualVariances(self._obj, columns=columns, alpha=alpha, **kwargs)
        
        equal_var = equal_variances

    @bluebelt.core.decorators.class_methods
    class graph():
        def __init__(self, pandas_obj):
            self._obj = pandas_obj
        
        def line(self, **kwargs):
            return bluebelt.analysis.graph.line(self._obj, **kwargs)

        def area(self, **kwargs):
            return bluebelt.analysis.graph.area(self._obj, **kwargs)

        def scatter(self, **kwargs):
            return bluebelt.analysis.graph.scatter(self._obj, **kwargs)

        def boxplot(self, **kwargs):
            return bluebelt.analysis.graph.boxplot(self._obj, **kwargs)



import pandas as pd
import numpy as np
import math
import scipy.stats as stats

import matplotlib.pyplot as plt
from matplotlib.ticker import FormatStrFormatter

import bluebelt.analysis.ci
import bluebelt.core.helpers
import bluebelt.styles
import bluebelt.core.decorators

from bluebelt.core.checks import check_kwargs

import warnings

def index():
    df = pd.DataFrame([['discrete', 'discrete', '-', '-', '-', 'chi square test', ''],
                        ['discrete', 'continuous', '-', '-', '-', 'logistic regression', ''],
                        ['continuous', 'discrete', 'mean', '1', 'normal', '1 sample t-test', 'one_sample_t'],
                        ['continuous', 'discrete', 'mean', '1', 'non-normal', '1 sample Wilcoxon test', 'wilcoxon'],
                        ['continuous', 'discrete', 'mean', '2', 'normal', '2 sample t-test', 'two_sample_t'],
                        ['continuous', 'discrete', 'mean', '2', 'non-normal', 'Mann-Whitney test', 'mann_whitney'],
                        ['continuous', 'discrete', 'mean', '>2', 'normal', '1 way Anova', 'anova'],
                        ['continuous', 'discrete', 'mean', '>2', 'non-normal', 'Kruskal-Wallis test', 'kruskal'],
                        ['continuous', 'discrete', 'variance', '2', 'normal', 'F-test', 'f_test'],
                        ['continuous', 'discrete', 'variance', '2', 'non-normal', 'Levene’s test', 'levene'],
                        ['continuous', 'discrete', 'variance', '>2', 'normal', 'Bartlett’s test', 'bartlett'],
                        ['continuous', 'discrete', 'variance', '>2', 'non-normal', 'Levene’s test', 'levene'],
                        ['continuous', 'continuous', '', '-', '-', 'regression', '']],
                    columns=['Y', 'X', 'investigate', '# data groups', 'distribution', 'test', 'bluebelt'])
    print(df.to_string(index=False))

# equal means
@bluebelt.core.decorators.class_methods
class OneSampleT():

    def __init__(self, frame, columns=None, popmean=None, confidence=0.95, alpha=0.05, **kwargs):

        if isinstance(frame, pd.Series):
            frame = pd.DataFrame(frame)
        elif isinstance(columns, str):
            frame = frame[columns]
        elif isinstance(columns, list):
            frame = frame[columns[0]]
        else:
            frame = frame[:,:1]

        # check arguments
        ncols = 1
        check_kwargs(locals())
                        
        self.frame = frame
        self.columms = columns
        self.nrows = frame.shape[0]
        self.ncols = frame.shape[1]
        self.popmean = popmean
        self.confidence = confidence
        self.alpha = alpha
        self.test = "One Sample T"
        self.h0 = f'$H_0: \\bar{{X}} = \\mu$'
        self.calculate(**kwargs)

    def calculate(self, **kwargs):
        self.statistic, self.p_value = stats.ttest_1samp(a=self.frame.iloc[:,0].values, popmean=self.popmean, **kwargs)
        self.ci_mean = [bluebelt.analysis.ci.ci_mean(self.frame.iloc[:,0])]
        self.passed = True if self.p_value > self.alpha else False

    def __str__(self):
        return ""
    
    def __repr__(self):
        return (f'{self.__class__.__name__}(nrows={self.nrows}, ncols={self.ncols}, alpha={self.alpha}, p_value={self.p_value:1.2f}, passed={self.passed})')
    
    def result(self):
        print(self.statistic)
    
    def plot(self, **kwargs):
        return _equal_means_plot(self, **kwargs)

@bluebelt.core.decorators.class_methods
class Wilcoxon():
    """
    Calculate the Wilcoxon signed-rank test.

    The Wilcoxon signed-rank test tests the null hypothesis that two related paired samples come from the same distribution.
    In particular, it tests whether the distribution of the differences x - y is symmetric about zero.
    It is a non-parametric version of the paired T-test.
    This test is performed with the assumption that if a pandas Series is provided the series contains the differences between x and y.
    """
    def __init__(self, frame, columns=None, confidence=0.95, alpha=0.05, **kwargs):

        if isinstance(frame, pd.Series):
            frame = pd.DataFrame(frame)
        elif isinstance(columns, str):
            frame = frame[columns]
        elif isinstance(columns, list):
            frame = frame[columns[:2]]
        else:
            frame = frame[:,:2]

        # check arguments
        ncols = (1, min(frame.shape[1], 2))
        check_kwargs(locals())
        
        self.frame = frame
        self.columms = columns
        self.nrows = self.frame.shape[0]
        self.ncols = self.frame.shape[1]
        self.confidence = confidence
        self.alpha = alpha
        self.test = "Wilcoxon signed-rank test"
        self.h0 = f'difference between the pairs follows a symmetric distribution around zero'
        self.calculate(**kwargs)

    def calculate(self, **kwargs):
        self.array_a = self.frame.iloc[:,0].dropna()
        self.array_b = self.frame.iloc[:,1].dropna() if self.ncols == 2 else None
        self.statistic, self.p_value = stats.wilcoxon(x=self.array_a, y=self.array_b, **kwargs)
        self.passed = True if self.p_value > self.alpha else False

    def __str__(self):
        return ""
    
    def __repr__(self):
        return (f'{self.__class__.__name__}(nrows={self.nrows}, ncols={self.ncols}, alpha={self.alpha}, p_value={self.p_value:1.2f}, passed={self.passed})')
    
    def plot(self, **kwargs):
        return _equal_means_plot(self, **kwargs)

@bluebelt.core.decorators.class_methods
class TwoSampleT():

    def __init__(self, frame, columns=None, related=False, confidence=0.95, alpha=0.05, **kwargs):
        
        if isinstance(columns, list):
            frame = frame[columns[:2]]
        else:
            frame = frame.iloc[:,:2]

        # check arguments
        ncols = 2
        check_kwargs(locals())
        
        self.frame = frame
        self.columms = columns
        self.nrows = self.frame.shape[0]
        self.ncols = self.frame.shape[1]
        self.related = related
        self.confidence = confidence
        self.alpha = alpha
        self.test = "Two Sample T Test"
        self.h0 = bluebelt.core.helpers._get_h0_equal_means(self.frame.columns)
        self.calculate(**kwargs)

    def calculate(self, **kwargs):
        self.array_a = self.frame.iloc[:,0].dropna()
        self.array_b = self.frame.iloc[:,1].dropna()
        
        self.mean_a = self.array_a.mean()
        self.mean_b = self.array_b.mean()
        
        self.max_a = self.array_a.max()
        self.max_b = self.array_b.max()
        
        self.min_a = self.array_a.min()
        self.min_b = self.array_b.min()
        
        self.ci_mean_a = bluebelt.analysis.ci.ci_mean(self.array_a)
        self.ci_mean_b = bluebelt.analysis.ci.ci_mean(self.array_b)

        # related or not
        if self.related:
            statistic, pvalue = stats.ttest_rel(self.array_a, self.array_b, **kwargs)
        else:
            # test equal variance
            equal_var = True if Levene(frame=self.frame, alpha=self.alpha).passed else False
            statistic, pvalue = stats.ttest_ind(self.array_a, self.array_b, equal_var=equal_var, **kwargs)

        self.statistic = statistic
        self.p_value = pvalue
        self.passed = True if self.p_value >= self.alpha else False


    def __str__(self):
        return ""
    
    def __repr__(self):
        return (f'{self.__class__.__name__}(nrows={self.nrows}, ncols={self.ncols}, alpha={self.alpha}, p_value={self.p_value:1.2f}, passed={self.passed})')
    
    def result(self):
        print(self.statistic)
    
    def plot(self, **kwargs):
        return _equal_means_plot(self, **kwargs)

@bluebelt.core.decorators.class_methods
class MannWhitney():
    """
    apply the Mann-Whitney test on a pandas dataframe with the columns as groups to compare
    Returns the scipy Mann-Whitney; statistic, pvalue
    pvalue >= alpha (default 0.05) : columns have equal means
    pvalue < alpha: columns don't have equal means
    """
    
    def __init__(self, frame, columns=None, confidence=0.95, alpha=0.05, **kwargs):
        
        # check arguments
        ncols = 2
        check_kwargs(locals())

        self.frame = frame if columns is None else frame[columns]
        self.columms = columns
        self.nrows = self.frame.shape[0]
        self.ncols = self.frame.shape[1]
        self.confidence = confidence
        self.alpha = alpha
        self.test = 'Mann-Whitney'
        self.h0 = bluebelt.core.helpers._get_h0_equal_means(self.frame.columns)
        self.calculate(**kwargs)

    def calculate(self, **kwargs):
        
        self.statistic, self.p_value = stats.mannwhitneyu(*self.frame.dropna().T.values)
        self.passed = True if self.p_value >= self.alpha else False

    def __repr__(self):
        return (f'{self.__class__.__name__}(nrows={self.nrows}, ncols={self.ncols}, alpha={self.alpha}, p_value={self.p_value:1.2f}, passed={self.passed})')
    
    def plot(self, **kwargs):
        return _equal_means_plot(self, **kwargs)

@bluebelt.core.decorators.class_methods
class Anova():
    """
    apply the one way ANOVA test on a pandas dataframe with the columns as groups to compare
    Returns the scipy F_onewayResult; statistic, pvalue
    pvalue >= alpha (default 0.05) : columns have equal means
    pvalue < alpha: columns don't have equal means
    """
    
    def __init__(self, frame, columns=None, confidence=0.95, alpha=0.05, **kwargs):
        
        # check arguments
        ncols = (2, None)
        check_kwargs(locals())

        self.frame = frame if columns is None else frame[columns]
        self.data_dict = _data_dict(self.frame)
        self.columms = columns
        self.nrows = self.frame.shape[0]
        self.ncols = self.frame.shape[1]
        self.confidence = confidence
        self.alpha = alpha
        self.test = "ANOVA"
        self.h0 = bluebelt.core.helpers._get_h0_equal_means(self.frame.columns)
        self.calculate(**kwargs)

    def calculate(self, **kwargs):
        
        self.statistic, self.p_value = stats.f_oneway(*self.data_dict.values())
        self.passed = True if self.p_value >= self.alpha else False

    def __repr__(self):
        return (f'{self.__class__.__name__}(nrows={self.nrows}, ncols={self.ncols}, alpha={self.alpha}, p_value={self.p_value:1.2f}, passed={self.passed})')
    
    def plot(self, **kwargs):
        return _equal_means_plot(self, **kwargs)

@bluebelt.core.decorators.class_methods
class KruskalWallis():
    """
    apply the Kruskal-Wallis test on a pandas dataframe with the columns as groups to compare
    Returns the scipy KruskalResult; statistic, pvalue
    pvalue >= alpha (default 0.05) : columns have equal means
    pvalue < alpha: columns don't have equal means
    """
    
    def __init__(self, frame, columns=None, confidence=0.95, alpha=0.05, **kwargs):
        
        # check arguments
        ncols = (2, None)
        check_kwargs(locals())


        self.frame = frame if columns is None else frame[columns]
        self.data_dict = _data_dict(self.frame)
        self.columms = columns
        self.nrows = self.frame.shape[0]
        self.ncols = self.frame.shape[1]
        self.confidence = confidence
        self.alpha = alpha
        self.test = 'Kruskal-Wallis'
        self.h0 = bluebelt.core.helpers._get_h0_equal_means(self.frame.columns)
        self.calculate(**kwargs)

    def calculate(self, **kwargs):
        
        self.statistic, self.p_value = stats.kruskal(*self.data_dict.values())
        self.passed = True if self.p_value >= self.alpha else False

    def __repr__(self):
        return (f'{self.__class__.__name__}(nrows={self.nrows}, ncols={self.ncols}, alpha={self.alpha}, p_value={self.p_value:1.2f}, passed={self.passed})')
    
    def plot(self, **kwargs):
        return _equal_means_plot(self, **kwargs)

@bluebelt.core.decorators.class_methods
class EqualMeans():
    """
    Test a dataframe for equal means over the dataframe columns.

    frame: a pandas DataFrame
    columns: string or list of strings with the column names to be tested
    alpha: test threshold

    applies ANOVA or Kruskal-Wallis test

    pvalue >= alpha (default 0.05) : columns have equal means
    pvalue < alpha: columns don't have equal means
    
    """
    
    def __init__(self, frame, columns=None, confidence=0.95, alpha=0.05, **kwargs):
        
       # check arguments
        ncols = (2, None)
        check_kwargs(locals())

        self.frame = frame if columns is None else frame[columns]
        self.columms = columns
        self.nrows = self.frame.shape[0]
        self.ncols = self.frame.shape[1]
        self.confidence = confidence
        self.alpha = alpha
        self.h0 = bluebelt.core.helpers._get_h0_equal_means(self.frame.columns)
        self.calculate(**kwargs)

    def calculate(self, **kwargs):
        
        # test for normal distribution; null hypothesis: values come from a normal distribution
        if NormalDistribution(frame=self.frame, alpha=self.alpha).passed and Levene(frame=self.frame, alpha=self.alpha).passed:
            # all columns in the dataframe come from a normal distribution AND have equal variances
            # do Anova
            test = Anova(frame=self.frame, alpha=self.alpha)
        else:
            # not all columns in the dataframe come from a normal distribution OR have equal variances
            # do Kruskal-Wallis
            test = KruskalWallis(frame=self.frame, alpha=self.alpha)
        
        self.test = test.test
        self.statistic = test.statistic
        self.p_value = test.p_value
        self.passed = True if self.p_value >= self.alpha else False

    def __repr__(self):
        return (f'{self.__class__.__name__}(nrows={self.nrows}, ncols={self.ncols}, alpha={self.alpha}, p_value={self.p_value:1.2f}, passed={self.passed})')
    
    def plot(self, **kwargs):
        return _equal_means_plot(self, **kwargs)

def _equal_means_plot(plot_obj, **kwargs):

    style = kwargs.pop('style', bluebelt.styles.paper)
    path = kwargs.pop('path', None)
    digits = kwargs.pop('digits', 3)
    xlim = kwargs.pop('xlim', None)

    # prepare figure
    fig, axes = plt.subplots(nrows=plot_obj.ncols, ncols=1, gridspec_kw={'wspace': 0, 'hspace': 0}, **kwargs)

    for row in range(0, plot_obj.ncols):

        if plot_obj.ncols == 1:
            ax = axes
        else:
            ax = axes[row]
        
        # set right zorder
        ax.set_zorder((plot_obj.ncols-row)*10)
        # get the data
        array = plot_obj.frame.iloc[:,row].dropna()

        # 1. box plot ############################################
        boxplot = ax.boxplot(array, vert=False, widths=0.3, whis=1.5)
        for box in boxplot['boxes']:
            box.set(**style.equal_means_plot.boxplot.get('boxes', {}))
        
        # 2. CI for the mean #####################################
        ci_mean = bluebelt.analysis.ci.ci_mean(array, confidence=plot_obj.confidence)

        ax.plot([ci_mean[0], ci_mean[1]], [1.5,1.5], **style.equal_means_plot.ci_mean_plot)
        ax.axvline(x=ci_mean[0], ymin=9/14, ymax=11/14, **style.equal_means_plot.ci_mean_axvline)
        ax.axvline(x=ci_mean[1], ymin=9/14, ymax=11/14, **style.equal_means_plot.ci_mean_axvline)
        ax.scatter([array.mean()],[1.5], **style.equal_means_plot.ci_mean_scatter)

        ci_area = np.linspace(ci_mean[0], ci_mean[1], 100)
        ax.fill_between(x=(ci_mean[0], ci_mean[1]), y1=0, y2=2, label='CI sample mean', **style.equal_means_plot.ci_fill_between)
        ax.axvline(x=ci_mean[0], ymin=0, ymax=2, **style.equal_means_plot.ci_axvline)
        ax.axvline(x=ci_mean[1], ymin=0, ymax=2, **style.equal_means_plot.ci_axvline)
        
        
        # plot CI values
        ci_digits_str = bluebelt.core.helpers._format_digits([ci_mean[0], ci_mean[1]], digits=digits)
        ax.text(ci_mean[0], 1.5, f'{ci_mean[0]:{ci_digits_str}} ', **style.equal_means_plot.text_ci_min)
        ax.text(ci_mean[1], 1.5, f' {ci_mean[1]:{ci_digits_str}}', **style.equal_means_plot.text_ci_max)

        # plot popmean if it exists
        if hasattr(plot_obj, 'popmean'):
            ax.axvline(plot_obj.popmean, ymin=0, ymax=2, label='population mean', **style.equal_means_plot.popmean_axvline)
        

        # set xticks
        local_xticks = bluebelt.core.helpers._get_boxplot_quantiles(array, whis=1.5)
        ax.set_xticks(local_xticks)
        ax.tick_params(axis="x", top=False, direction='in', pad=-18)
        

        ax.set_ylim([0.25, 2])
        ax.set_yticklabels([array.name])

        # add xticks to xtick list for formatting
        if row == 0:
            xticks = local_xticks
            xlims = ax.get_xlim()
            
            ax.set_title(plot_obj.h0 + f'   t={plot_obj.statistic:1.2f} p={plot_obj.p_value:1.2f} ('+r'$H_0$'+f' is {"accepted" if plot_obj.p_value > plot_obj.alpha else "rejected"})', **style.equal_means_plot.title)

            ax.legend(loc='best')

        else:
            xticks += local_xticks
            xlims = (min(ax.get_xlim()[0], xlims[0]), max(ax.get_xlim()[1], xlims[1]))
        
    # format things
    for row in range(0, plot_obj.ncols):

        if plot_obj.ncols == 1:
            ax = axes
        else:
            ax = axes[row]

        ax.xaxis.set_major_formatter(FormatStrFormatter(f'%{bluebelt.core.helpers._format_digits(xticks, digits=digits)}'))
        if xlim is not None:
            ax.set_xlim(xlim)
        else:
            ax.set_xlim(xlims)
        

    # add suptitle
    fig.suptitle(t=f'{plot_obj.test} for {bluebelt.core.helpers._get_nice_list(plot_obj.frame.columns)}', x=fig.subplotpars.left, **style.equal_means_plot.suptitle)

    if path:
        plt.savefig(path)
        plt.close()
    else:
        plt.close()
        return fig

# equal variance
@bluebelt.core.decorators.class_methods
class FTest():
    """
    apply the F-test on a pandas dataframe with the columns as groups to compare
    the F-test is extremely sensitive to non-normality of the two arrays
    pvalue >= 0.05 : columns have equal variances
    pvalue < 0.05: columns don't have equal variances
    """
    
    def __init__(self, frame, columns=None, alpha=0.05, **kwargs):
        
        # check arguments
        ncols = 2
        check_kwargs(locals())
        
        self.frame = frame if columns is None else frame[columns]
        self.columms = columns
        self.array_a = self.frame.iloc[:,0].dropna()
        self.array_b = self.frame.iloc[:,1].dropna()
        self.nrows = self.frame.shape[0]
        self.ncols = self.frame.shape[1]
        self.alpha = alpha
        self.test = 'F-test'
        self.calculate(**kwargs)

    def calculate(self, **kwargs):
        self.statistic = self.array_a.var() / self.array_b.var()
        self.p_value = stats.f.cdf(self.statistic, self.array_a.size-1, self.array_b.size-1)
        self.passed = True if self.p_value >= self.alpha else False

    def __repr__(self):
        return (f'{self.__class__.__name__}(nrows={self.nrows}, ncols={self.ncols}, alpha={self.alpha}, p_value={self.p_value:1.2f}, passed={self.passed})')
    
    def plot(self, **kwargs):
        return _equal_variances_plot(self, **kwargs)

@bluebelt.core.decorators.class_methods
class Levene():
    """
    apply levene's test on a pandas dataframe with the columns as groups to compare
    Returns the scipy LeveneResult; statistic, pvalue
    pvalue >= 0.05 : columns have equal variances
    pvalue < 0.05: columns don't have equal variances
    """
    
    def __init__(self, frame, columns=None, alpha=0.05, **kwargs):
        
        # check arguments
        ncols = (2, None)
        check_kwargs(locals())
        

        self.frame = frame if columns is None else frame[columns]
        self.data_dict = _data_dict(self.frame)
        self.columms = columns
        self.nrows = self.frame.shape[0]
        self.ncols = self.frame.shape[1]
        self.alpha = alpha
        self.test = 'Levene'
        self.calculate(**kwargs)

    def calculate(self, **kwargs):
        self.statistic, self.p_value = stats.levene(*self.data_dict.values())
        self.passed = True if self.p_value >= self.alpha else False
    
    def __repr__(self):
        return (f'{self.__class__.__name__}(nrows={self.nrows}, ncols={self.ncols}, alpha={self.alpha}, p_value={self.p_value:1.2f}, passed={self.passed})')
    
    def plot(self, **kwargs):
        return _equal_variances_plot(self, **kwargs)

@bluebelt.core.decorators.class_methods
class Bartlett():
    """
    apply Bartlett's test on a pandas dataframe with the columns as groups to compare
    Returns the scipy BartlettResult; statistic, pvalue
    pvalue >= 0.05 : columns have equal variances
    pvalue < 0.05: columns don't have equal variances
    """
    
    def __init__(self, frame, columns=None, alpha=0.05, **kwargs):
        
        # check arguments
        ncols = (2,None)
        check_kwargs(locals())
        
        self.frame = frame if columns is None else frame[columns]
        self.data_dict = _data_dict(self.frame)
        self.columms = columns
        self.nrows = self.frame.shape[0]
        self.ncols = self.frame.shape[1]
        self.alpha = alpha
        self.test = 'Bartlett'
        self.calculate(**kwargs)

    def calculate(self, **kwargs):
        self.statistic, self.p_value = stats.bartlett(*self.data_dict.values())
        self.passed = True if self.p_value >= self.alpha else False
    
    def __repr__(self):
        return (f'{self.__class__.__name__}(nrows={self.nrows}, ncols={self.ncols}, alpha={self.alpha}, p_value={self.p_value:1.2f}, passed={self.passed})')
    
    def plot(self, **kwargs):
        return _equal_variances_plot(self, **kwargs)

@bluebelt.core.decorators.class_methods
class EqualVariances():
    """
    Test a dataframe for equal variances over the dataframe columns.

    frame: a pandas DataFrame
    columns: string or list of strings with the column names to be tested
    alpha: test threshold

    applies F-test, Bartlett's or Levene's

    pvalue >= alpha (default 0.05) : columns have equal variances
    pvalue < alpha: columns don't have equal variances
    
    """
    def __init__(self, frame, columns=None, alpha=0.05, **kwargs):
        
        # check arguments
        ncols = (2,None)
        check_kwargs(locals())
        
        self.frame = frame if columns is None else frame[columns]
        self.columms = columns
        self.nrows = self.frame.shape[0]
        self.ncols = self.frame.shape[1]
        self.alpha = alpha
        self.calculate(**kwargs)

    def calculate(self, **kwargs):

        # test for normal distribution; null hypothesis: values come from a normal distribution
        if NormalDistribution(frame=self.frame, alpha=self.alpha).passed:
            # all columns in the dataframe come from a normal distribution
            if self.ncols == 2:
                test = FTest(frame=self.frame, alpha=self.alpha)
            else:
                test = Bartlett(frame=self.frame, alpha=self.alpha)
        else:
            # not all columns in the dataframe come from a normal distribution
            test=Levene(frame=self.frame, alpha=self.alpha)
        
        self.test = test.test
        self.statistic = test.statistic
        self.p_value = test.p_value
        self.passed = True if self.p_value >= self.alpha else False
    
    def __repr__(self):
        return (f'{self.__class__.__name__}(nrows={self.nrows}, ncols={self.ncols}, alpha={self.alpha}, p_value={self.p_value:1.2f}, passed={self.passed})')
    
    def plot(self, **kwargs):
        return _equal_variances_plot(self, **kwargs)

def _equal_variances_plot(plot_obj, **kwargs):

    style = kwargs.pop('style', bluebelt.styles.paper)
    path = kwargs.pop('path', None)
    digits = kwargs.pop('digits', 3)
    bins = kwargs.pop('bins', 17)

    # prepare figure
    fig, axes = plt.subplots(nrows=plot_obj.ncols, ncols=1, sharex='col', gridspec_kw={'wspace': 0, 'hspace': 0}, **kwargs)

    for row in range(0, plot_obj.ncols):

        if plot_obj.ncols == 1:
            ax = axes
        else:
            ax = axes[row]
        
        # set right zorder
        ax.set_zorder((plot_obj.ncols-row)*10)
        # get the data
        array = plot_obj.frame.iloc[:,row]

        # calculate bin width
        bin_width = (np.nanmax(plot_obj.frame.values) - np.nanmin(plot_obj.frame.values)) / bins

        # 1. histogram ############################################
        ax.hist(array, bins=np.arange(np.nanmin(plot_obj.frame.values), np.nanmax(plot_obj.frame.values) + bin_width, bin_width), **style.equal_variances_plot.histogram)
        ax.text(0.05, 0.9, array.name, transform=ax.transAxes, **style.equal_variances_plot.text)
        ax.set_yticks([])

        # add xlims to xlims list for formatting
        if row == 0:
            xlims = ax.get_xlim()
            ax.set_title(f't={plot_obj.statistic:1.2f} p={plot_obj.p_value:1.2f} ('+r'$H_0$'+f' is {"accepted" if plot_obj.p_value > plot_obj.alpha else "rejected"})', **style.equal_variances_plot.title)
 
        else:
            xlims = (min(ax.get_xlim()[0], xlims[0]), max(ax.get_xlim()[1], xlims[1]))
        
    # format things
    for row in range(0, plot_obj.ncols):

        if plot_obj.ncols == 1:
            ax = axes
        else:
            ax = axes[row]

        ax.set_xlim(xlims)

    # add suptitle
    fig.suptitle(t=f'{plot_obj.test} for {bluebelt.core.helpers._get_nice_list(plot_obj.frame.columns)}', x=fig.subplotpars.left, **style.equal_variances_plot.suptitle)
    
    if path:
        plt.savefig(path)
        plt.close()
    else:
        plt.close()
        return fig

# distribution
@bluebelt.core.decorators.class_methods
class AndersonDarling():
    """
    Anderson-Darling test
    
    null hypothesis: x comes from a normal distribution
    
    test all pandas dataframe columns for normal distribution
    input pivoted dataframe
    
    pvalue >= alpha : column values are (all) normally distributed
    pvalue < alpha: column values are not (all) normally distributed
    """
    
    def __init__(self, frame, columns=None, dist='norm', alpha=0.05, **kwargs):
        
        if isinstance(frame, pd.Series):
            frame = pd.DataFrame(frame)
        elif isinstance(columns, (list, str)):
            frame = frame[columns]
        
        # check arguments
        ncols = (1, None)
        check_kwargs(locals())

        self.frame = frame
        self.columns = columns
        self.dist = dist
        self.nrows = self.frame.shape[0]
        self.ncols = self.frame.shape[1]
        self.alpha = alpha
        self.test = "Anderson-Darling"
        self.calculate(**kwargs)

    def calculate(self, **kwargs):
        statistics = {}
        p_values = {}
        passed_values = {}

        statistic = 0
        p_value = 1
        
        for col in self.frame.columns:

            AD, critical_values, significance_level = stats.anderson(self.frame[col].dropna().values, dist=self.dist)
        
            AD_adjusted = AD*(1 + (.75/50) + 2.25/(50**2))
            if AD_adjusted >= .6:
                pval = math.exp(1.2937 - 5.709*AD_adjusted - .0186*(AD_adjusted**2))
            elif AD_adjusted >=.34:
                pval = math.exp(.9177 - 4.279*AD_adjusted - 1.38*(AD_adjusted**2))
            elif AD_adjusted >.2:
                pval = 1 - math.exp(-8.318 + 42.796*AD_adjusted - 59.938*(AD_adjusted**2))
            else:
                pval = 1 - math.exp(-13.436 + 101.14*AD_adjusted - 223.73*(AD_adjusted**2))

            if pval < p_value:
                p_value = pval
                statistic = AD
            
            statistics[col] = AD
            p_values[col] = pval
            passed_values[col] = True if pval >= self.alpha else False

        self.statistics = pd.Series(statistics)
        self.p_values = pd.Series(p_values)
        self.passed_values = pd.Series(passed_values)
        
        self.statistic = statistic
        self.p_value = p_value
        self.passed = True if self.p_value >= self.alpha else False

    def plot(self, **kwargs):
        return _distribution_plot(self, dist=self.dist, **kwargs)

@bluebelt.core.decorators.class_methods
class DAgostinoPearson():
    """
    D’Agostino-Pearson test
    
    null hypothesis: x comes from a normal distribution
    
    test all pandas dataframe columns for normal distribution
    input pivoted dataframe
    Returns the scipy NormaltestResult; statistic, pvalue for a single column
    Returns the worst scipy NormaltestResult as a tuple; statistic, pvalue for multiple columns
    pvalue >= alpha : column values are (all) normally distributed
    pvalue < alpha: column values are not (all) normally distributed
    """
    
    def __init__(self, frame, columns=None, alpha=0.05, **kwargs):
        
        if isinstance(frame, pd.Series):
            frame = pd.DataFrame(frame)
        elif isinstance(columns, (list, str)):
            frame = frame[columns]
        
        # check arguments
        ncols = (1, None)
        check_kwargs(locals())

        self.frame = frame
        self.columns = columns
        self.dist = 'norm'
        self.nrows = self.frame.shape[0]
        self.ncols = self.frame.shape[1]
        self.alpha = alpha
        self.test = "D’Agostino-Pearson"
        self.calculate(**kwargs)

    def calculate(self, **kwargs):
        statistics = {}
        p_values = {}
        passed_values = {}

        statistic = 0
        p_value = 1
        
        for col in self.frame.columns:
            result = stats.normaltest(self.frame[col].dropna().values)
            if result.pvalue < p_value:
                p_value = result.pvalue
                statistic = result.statistic

            statistics[col] = result.statistic
            p_values[col] = result.pvalue
            passed_values[col] = True if result.pvalue >= self.alpha else False

        self.statistics = pd.Series(statistics)
        self.p_values = pd.Series(p_values)
        self.passed_values = pd.Series(passed_values)

        self.statistic = statistic
        self.p_value = p_value
        self.passed = True if self.p_value >= self.alpha else False

    def plot(self, **kwargs):
        return _distribution_plot(self, dist='norm', **kwargs)

NormalDistribution = DAgostinoPearson

def _distribution_plot(plot_obj, **kwargs):

    style = kwargs.pop('style', bluebelt.styles.paper)
    path = kwargs.pop('path', None)
    dist = kwargs.pop('dist', 'norm')


    # prepare figure
    ncols = math.ceil(math.sqrt(plot_obj.ncols))
    nrows = math.ceil(plot_obj.ncols/ncols)
    # nrows = math.ceil(math.sqrt(plot_obj.ncols))
    # ncols = math.ceil(plot_obj.nrows/nrows)
    
    fig, axes = plt.subplots(nrows=nrows, ncols=ncols, gridspec_kw={'wspace': 0, 'hspace': 0}, **kwargs)

    for row in range(0, nrows):
        for col in range(0, ncols):
            id = col + (ncols * row)

            if nrows == 1:
                if ncols == 1:
                    ax = axes
                else:
                    ax = axes[col]
            else:
                ax = axes[row, col]

                
            # get the data
            if id < plot_obj.ncols:
                array = plot_obj.frame.iloc[:,id]
                col = plot_obj.frame.columns[id]
                
                # H0 the two distributions are identical, F(x)=G(x)
                parameters = eval("stats."+dist+".fit(array.values)");
                (osm, osr), (slope, intercept, r) = stats.probplot(array, dist=eval("stats."+dist), sparams=parameters, fit=True)

                # plot
                ax.scatter(osm, osr, **style.distribution_plot.scatter)
                ax.plot(osm, osm*slope + intercept, **style.distribution_plot.plot)
                
                ax.text(0.99, 0.01, f'statistic: {plot_obj.statistics[col]:1.2f}\np-value: {plot_obj.p_values[col]:1.2f}', transform=ax.transAxes, **style.distribution_plot.stat_text)
                
                if plot_obj.ncols > 1:
                    ax.text(0.01, 0.99, f'{plot_obj.frame.columns[id]}', transform=ax.transAxes, **style.distribution_plot.array_text)
                
                if id == 0:
                    ax.set_title(f'{plot_obj.test} for {bluebelt.core.helpers._get_distribution(plot_obj.dist)}', **style.distribution_plot.title)
                ax.set_xticks([])
                ax.set_yticks([])
                
                ax.set_ylim(ax.get_xlim())
                
            else:
                ax.axis('off')
            
    # add suptitle
    fig.suptitle(t=f'Probability plot of {bluebelt.core.helpers._get_nice_list(plot_obj.frame.columns)}', x=fig.subplotpars.left, **style.distribution_plot.suptitle)

    if path:
        plt.savefig(path)
        plt.close()
    else:
        plt.close()
        return fig

def _data_dict(frame):
    return {k:v.dropna().values for k, v in frame.items()}


import bluebelt.styles.paper.summary
import bluebelt.styles.paper.control_chart
import bluebelt.styles.paper.run_chart
import bluebelt.styles.paper.process_capability
import bluebelt.styles.paper.patterns
import bluebelt.styles.paper.boxplot
import bluebelt.styles.paper.rolling_std
import bluebelt.styles.paper.correlation

import bluebelt.styles.paper.equal_means_plot
import bluebelt.styles.paper.equal_variances_plot
import bluebelt.styles.paper.distribution_plot
import bluebelt.styles.paper.resolution
import bluebelt.styles.paper.forecast
import bluebelt.styles.paper.planning

import bluebelt.styles.paper.graphs


import pandas as pd
import numpy as np
import scipy.stats as stats


import bluebelt.datetime.dt

import bluebelt.statistics.std

import bluebelt.data.index
import bluebelt.data.resolution

import bluebelt.analysis.ci
import bluebelt.statistics.hypothesis_testing
import bluebelt.analysis.pattern
import bluebelt.analysis.forecast
import bluebelt.analysis.planning
import bluebelt.analysis.datetime
import bluebelt.analysis.graph
import bluebelt.analysis.performance

import bluebelt.core.decorators

@bluebelt.core.decorators.class_methods
@pd.api.extensions.register_series_accessor('blue')
class SeriesToolkit():
    def __init__(self, pandas_obj):
        self._obj = pandas_obj
        self.statistics = self.statistics(self._obj)
        self.pattern = self.pattern(self._obj)
        self.forecast = self.forecast(self._obj)
        self.planning = self.planning(self._obj)
        self.datetime = self.datetime(self._obj)
        self.data = self.data(self._obj)
        self.graph = self.graph(self._obj)
        self.test = self.test(self._obj)
        self.performance = self.performance(self._obj)

    # first level functions
    @bluebelt.core.decorators.add_docstring(bluebelt.data.index.iso_index)
    def iso_index(self, **kwargs):
        return bluebelt.data.index.iso_index(self._obj, **kwargs)

    @bluebelt.core.decorators.add_docstring(bluebelt.data.index.dt_index)
    def dt_index(self, **kwargs):
        return bluebelt.data.index.dt_index(self._obj, **kwargs)

    @bluebelt.core.decorators.add_docstring(bluebelt.data.resolution.GroupByDatetimeIndex)
    def group_index(self, **kwargs):
        return bluebelt.data.resolution.GroupByDatetimeIndex(self._obj, **kwargs)
    
    @bluebelt.core.decorators.add_docstring(bluebelt.analysis.graph.line)
    def line(self, **kwargs):
        return bluebelt.analysis.graph.line(self._obj, **kwargs)
     
    @bluebelt.core.decorators.class_methods
    class statistics():

        def __init__(self, pandas_obj):
            self._obj = pandas_obj

        # rolling std
        @bluebelt.core.decorators.add_docstring(bluebelt.statistics.std.RollingStd)
        def rolling_std(self, window=7, center=True, **kwargs):
            return bluebelt.statistics.std.RollingStd(self._obj, window=window, center=center, **kwargs)
                
        # standard deviation
        @bluebelt.core.decorators.add_docstring(bluebelt.statistics.std.StdWithin)
        def std_within(self, how=None, observations=2, **kwargs):
            return bluebelt.statistics.std.StdWithin(self._obj, how=how, observations=observations, **kwargs)        

    @bluebelt.core.decorators.class_methods
    class pattern():

        def __init__(self, pandas_obj):
            self._obj = pandas_obj
                
        # patterns
        @bluebelt.core.decorators.add_docstring(bluebelt.analysis.pattern.Polynomial)
        def polynomial(self, **kwargs):
            return bluebelt.analysis.pattern.Polynomial(self._obj, **kwargs)

        @bluebelt.core.decorators.add_docstring(bluebelt.analysis.pattern.Periodical)
        def periodical(self, **kwargs):
            return bluebelt.analysis.pattern.Periodical(self._obj, **kwargs)

    @bluebelt.core.decorators.class_methods
    class forecast():

        def __init__(self, pandas_obj):
            self._obj = pandas_obj
        
        @bluebelt.core.decorators.add_docstring(bluebelt.analysis.forecast.MAPE)
        def MAPE(self, **kwargs):
            return bluebelt.analysis.forecast.MAPE(self._obj, **kwargs)

        @bluebelt.core.decorators.add_docstring(bluebelt.analysis.forecast.WMAPE)
        def WMAPE(self, **kwargs):
            return bluebelt.analysis.forecast.WMAPE(self._obj, **kwargs)

        @bluebelt.core.decorators.add_docstring(bluebelt.analysis.forecast.SMAPE)
        def SMAPE(self, **kwargs):
            return bluebelt.analysis.forecast.SMAPE(self._obj, **kwargs)

        @bluebelt.core.decorators.add_docstring(bluebelt.analysis.forecast.MDA)
        def MDA(self, **kwargs):
            return bluebelt.analysis.forecast.MDA(self._obj, **kwargs)

        @bluebelt.core.decorators.add_docstring(bluebelt.analysis.forecast.MPE)
        def MPE(self, **kwargs):
            return bluebelt.analysis.forecast.MPE(self._obj, **kwargs)

        mape = MAPE
        wmape = WMAPE
        wMAPE = WMAPE
        smape = SMAPE
        mda = MDA
        mpe = MPE

    @bluebelt.core.decorators.class_methods
    class planning():

        def __init__(self, pandas_obj):
            self._obj = pandas_obj
        
        @bluebelt.core.decorators.add_docstring(bluebelt.analysis.forecast.MAPE)
        def effort(self, **kwargs):
            return bluebelt.analysis.planning.Effort(self._obj, **kwargs)

    @bluebelt.core.decorators.class_methods
    class datetime():
        def __init__(self, pandas_obj):
            self._obj = pandas_obj

        # week day
        @bluebelt.core.decorators.add_docstring(bluebelt.analysis.datetime.WeekDay)
        def week_day(self, **kwargs):
            return bluebelt.analysis.datetime.WeekDay(self._obj, **kwargs)

        weekday = week_day

        # day of the month
        @bluebelt.core.decorators.add_docstring(bluebelt.analysis.datetime.MonthDay)
        def month_day(self, **kwargs):
            return bluebelt.analysis.datetime.MonthDay(self._obj, **kwargs)

        day = month_day

        # week of the year
        @bluebelt.core.decorators.add_docstring(bluebelt.analysis.datetime.Week)
        def week(self, **kwargs):
            return bluebelt.analysis.datetime.Week(self._obj, **kwargs)

        # month of the year
        @bluebelt.core.decorators.add_docstring(bluebelt.analysis.datetime.Month)
        def month(self, **kwargs):
            return bluebelt.analysis.datetime.Month(self._obj, **kwargs)

        # year
        @bluebelt.core.decorators.add_docstring(bluebelt.analysis.datetime.Year)
        def year(self, **kwargs):
            return bluebelt.analysis.datetime.Year(self._obj, **kwargs) 

    @bluebelt.core.decorators.class_methods
    class data():

        def __init__(self, pandas_obj):
            self._obj = pandas_obj
                
        @bluebelt.core.decorators.add_docstring(bluebelt.data.resolution.GroupByDatetimeIndex)
        def group_index(self, **kwargs):
            return bluebelt.data.resolution.GroupByDatetimeIndex(self._obj, **kwargs)
        
        
    @bluebelt.core.decorators.class_methods
    class graph():
        def __init__(self, pandas_obj):
            self._obj = pandas_obj
        
        @bluebelt.core.decorators.add_docstring(bluebelt.analysis.graph.line)
        def line(self, **kwargs):
            return bluebelt.analysis.graph.line(self._obj, **kwargs)

        @bluebelt.core.decorators.add_docstring(bluebelt.analysis.graph.scatter)
        def scatter(self, **kwargs):
            return bluebelt.analysis.graph.scatter(self._obj, **kwargs)

        @bluebelt.core.decorators.add_docstring(bluebelt.analysis.graph.area)
        def area(self, **kwargs):
            return bluebelt.analysis.graph.area(self._obj, **kwargs)

        @bluebelt.core.decorators.add_docstring(bluebelt.analysis.graph.hist)
        def hist(self, **kwargs):
            return bluebelt.analysis.graph.hist(self._obj, **kwargs)
            
        histogram = hist

        @bluebelt.core.decorators.add_docstring(bluebelt.analysis.graph.boxplot)
        def boxplot(self, **kwargs):
            return bluebelt.analysis.graph.boxplot(self._obj, **kwargs)
            
    @bluebelt.core.decorators.class_methods
    class test():
        def __init__(self, pandas_obj):
            self._obj = pandas_obj

        # index
        @property
        def index(self):
            return bluebelt.statistics.hypothesis_testing.index()

        # hypothesis testing
        @bluebelt.core.decorators.add_docstring(bluebelt.statistics.hypothesis_testing.NormalDistribution)
        def normal_distribution(self, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.NormalDistribution(self._obj, alpha=alpha)
        
        @bluebelt.core.decorators.add_docstring(bluebelt.statistics.hypothesis_testing.DAgostinoPearson)
        def dagostino_pearson(self, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.DAgostinoPearson(self._obj, alpha=alpha)
        
        @bluebelt.core.decorators.add_docstring(bluebelt.statistics.hypothesis_testing.AndersonDarling)
        def anderson_darling(self, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.AndersonDarling(self._obj, alpha=alpha)
                
        @bluebelt.core.decorators.add_docstring(bluebelt.statistics.hypothesis_testing.OneSampleT)
        def one_sample_t(self, popmean=None, confidence=0.95, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.OneSampleT(self._obj, popmean=popmean, confidence=confidence, alpha=alpha, **kwargs)

        @bluebelt.core.decorators.add_docstring(bluebelt.statistics.hypothesis_testing.Wilcoxon)
        def wilcoxon(self, alpha=0.05, **kwargs):
            return bluebelt.statistics.hypothesis_testing.Wilcoxon(self._obj, alpha=alpha, **kwargs)
        
    @bluebelt.core.decorators.class_methods
    class performance():
        def __init__(self, pandas_obj):
            self._obj = pandas_obj

        @bluebelt.core.decorators.add_docstring(bluebelt.analysis.performance.Summary)
        def summary(self, **kwargs):
            return bluebelt.analysis.performance.Summary(self._obj, **kwargs)
        
        @bluebelt.core.decorators.add_docstring(bluebelt.analysis.performance.ControlChart)
        def control_chart(self, **kwargs):
            return bluebelt.analysis.performance.ControlChart(self._obj, **kwargs)
            
        @bluebelt.core.decorators.add_docstring(bluebelt.analysis.performance.RunChart)
        def run_chart(self, alpha=0.05, **kwargs):
            return bluebelt.analysis.performance.RunChart(self._obj, alpha=alpha, **kwargs)

        @bluebelt.core.decorators.add_docstring(bluebelt.analysis.performance.ProcessCapability)
        def process_capability(self, **kwargs):
            return bluebelt.analysis.performance.ProcessCapability(self._obj, **kwargs)
        
        capability = process_capability

        pca = process_capability


import bluebelt.styles.defaults as defaults

# main plot | pyplot.plot
plot = {
    #'color': defaults.black,
    'linestyle': 'solid',
    'linewidth': 0,
    'marker': 'o',
    'markersize': 7,
    'markerfacecolor': defaults.blue,
    'markeredgewidth': 0,
    'markeredgecolor': defaults.blue,
    'alpha': 0.7,
    'zorder': 90,
}

# plot title | pyplot.set_title
title = {
    'loc': 'left',
}


import pandas as pd
import numpy as np
import scipy.stats as stats

import warnings

import matplotlib.pyplot as plt

import bluebelt.core.helpers
import bluebelt.core.decorators

import bluebelt.data.resolution

import bluebelt.styles

@bluebelt.core.decorators.class_methods


class Effort():
    """
    Calculate the planning effort
    """
    def __init__(self, series, how='group', rule='iso_jaar-iso_week', **kwargs):
        
        self.series = series
        self.how = how
        self.rule = rule

        # get qdq variable columns
        self.quantity = kwargs.pop('quantity', None) or kwargs.pop('q', None)
        self.distribution = kwargs.pop('distribution', None) or kwargs.pop('dist', None) or kwargs.pop('d', None)
        self.quality = kwargs.pop('quality', None) or kwargs.pop('skills', None) or kwargs.pop('s', None)
        
        print(self.quantity)
        print(self.distribution)
        print(self.quality)
        
        self.calculate()

    def calculate(self):

        if self.how == 'group':
            self.quantity = bluebelt.data.resolution.GroupByDatetimeIndex(self.series, rule=self.rule).inter_scale().result
            self.distribution = bluebelt.data.resolution.GroupByDatetimeIndex(self.series, rule=self.rule).intra_scale().result
        

    def __repr__(self):
        return (f'{self.__class__.__name__}(n={self.series.size:1.0f}, quantity={self.quantity.mean():1.4f}, distribution={self.distribution.mean():1.4f})')
    
    def plot(self, **kwargs):
        
        return _qdq_plot(self, **kwargs)

def _qdq_plot(_obj, **kwargs):
        
    style = kwargs.pop('style', bluebelt.styles.paper)
    title = kwargs.pop('title', f'{_obj.series.name} Effort QDQ plot')
    
    path = kwargs.pop('path', None)
    xlim = kwargs.pop('xlim', (None, None))
    ylim = kwargs.pop('ylim', (0, None))
    
    # prepare figure
    fig, axes = plt.subplots(nrows=1, ncols=1, **kwargs)

    axes.fill_between(_obj.quantity.index, 0, _obj.quantity.values, **style.planning.fill_between_inter, label='quantity')
    axes.plot(_obj.inter, **style.planning.plot_quantity, **kwargs)

    axes.fill_between(_obj.distribution.index, 0, _obj.distribution.values, **style.planning.fill_between_intra, label='distribution')
    axes.plot(_obj.intra, **style.planning.plot_distribution, **kwargs)

    # format things
    axes.set_xlim(xlim)
    axes.set_ylim(ylim)
    
    # title
    axes.set_title(title, **style.graphs.line.title)

    # legend
    axes.legend(loc='upper left')

    if path:
        plt.savefig(path)
        plt.close()
    else:
        plt.close()
        return fig


import pandas as pd
import numpy as np

def check_kwargs(kwargs):


    kwargs = _flatten_dict(kwargs)

    series = kwargs.pop('series', None)
    frame = kwargs.pop('frame', None)
    columns = kwargs.pop('columns', None)
    ncols = kwargs.pop('ncols', None)
    alpha = kwargs.pop('alpha', None)
    confidence = kwargs.pop('confidence', None)
    
    
    if series is not None and not isinstance(series, pd.Series):
        raise ValueError(f'series must be a pandas Series not a {type(series)}')

    if isinstance(frame, pd.DataFrame):
        if columns is not None and not isinstance(columns, (str, list)):
            raise ValueError(f'columns must be a string or a list not a {type(columns)}')
            
        if ncols is not None and not isinstance(ncols, (int, list, tuple)):
            # this should never happen
            raise ValueError(f'ncols must be an integer, list or tuple not a {type(ncol)}')
        
        if isinstance(ncols, int):
            if isinstance(columns, str) and ncols != 1:
                raise ValueError(f'columns must be a list not a {type(columns)}')
            elif isinstance(columns, list) and ncols != len(columns):
                raise ValueError(f'columns must be a list of {ncols} columns not {len(columns)} columns')
            elif columns is None and frame.shape[1] != ncols:
                raise ValueError(f'frame must have {ncols} columns not {frame.shape[1]} columns')
        elif isinstance(ncols, (list, tuple)):
            if ncols[0] is None:
                # column length must be less than or equal to ncols[1}]
                if isinstance(columns, str) and 1 != ncols[1]:
                    raise ValueError(f'columns must be a list not a {type(columns)}')
                if isinstance(columns, list) and len(columns) > ncols[1]:
                    raise ValueError(f'columns must be a list of {ncols[1]} or less columns not {len(columns)} columns')
                elif frame.shape[1] > ncols[1]:
                    raise ValueError(f'frame must have {ncols[1]} or less columns not {frame.shape[1]} columns')
            elif ncols[1] is None:
                # column length must be greater than or equal to ncols[1}]
                if isinstance(columns, str) and 1 != ncols[0]:
                    raise ValueError(f'columns must be a list not a {type(columns)}')
                if isinstance(columns, list) and len(columns) < ncols[0]:
                    raise ValueError(f'columns must be a list of {ncols[0]} or more columns not {len(columns)} columns')
                elif frame.shape[1] < ncols[0]:
                    raise ValueError(f'frame must have {ncols[0]} or more columns not {frame.shape[1]} columns')
            else:
                # column length must be between ncols[0] and ncols[1]
                if isinstance(columns, str) and 1 not in np.arange(ncols[0], ncols[1]+1):
                    raise ValueError(f'columns must be a list not a {type(columns)}')
                if isinstance(columns, list) and len(columns) not in np.arange(ncols[0], ncols[1]+1):
                    raise ValueError(f'columns must be a list of between {ncols[0]} and {ncols[1]} columns not {len(columns)} columns')
                elif frame.shape[1] not in np.arange(ncols[0], ncols[1]+1):
                    raise ValueError(f'frame must have between {ncols[0]} and {ncols[1]} columns not {frame.shape[1]} columns')
    elif frame is not None:
        raise ValueError(f'frame must be a pandas Series or a pandas DataFrame not a {type(frame)}')

    if alpha is not None and not isinstance(alpha, (int, float)):
        raise ValueError(f'alpha must be a float or an integer not a {type(alpha)}')
    elif alpha is not None and not (0 < alpha < 1):
        raise ValueError(f'0 < alpha < 1 is required, but alpha={alpha} was given')
    
    if confidence is not None and not (0 < confidence < 1):
        raise ValueError(f'0 < confidence < 1 is required, but confidence={confidence} was given')
    
    if 'popmean' in kwargs:
        popmean = kwargs.pop('popmean', None)
        if popmean is None:
            raise ValueError('popmean must be provided')
        elif not isinstance(popmean, (int, float)):
            raise ValueError(f'popmean must be an integer or a float not a {type(popmean)}')

    # return rest kwargs
    return kwargs

def _flatten_dict(d):
    def expand(key, value):
        if isinstance(value, dict):
            return [ (k, v) for k, v in _flatten_dict(value).items() ]
        else:
            return [ (key, value) ]

    items = [ item for k, v in d.items() for item in expand(k, v) ]

    return dict(items)



import pandas as pd
import scipy.stats as stats

import matplotlib.pyplot as plt

import bluebelt.statistics.hypothesis_testing

import bluebelt.styles

class WeekDay(): 
    """
    Compare the distribution of data between week days
        arguments
        series: pandas.Series
            the Series must have a pandas.DatetimeIndex
        
        properties
        .series
            the transformed pandas.Series with week day index
        .data
            the data
        .equal_means
            the result of bluebelt.statistics.hypothesis_testing.EqualMeans().passed
        .equal_variances
            the result of bluebelt.statistics.hypothesis_testing.EqualVariances().passed

        methods
        .plot()
            plot the results as a boxplot
    """
    

    def __init__(self, series, **kwargs):
        
        self.series = series
        self.name = series.name
        self.calculate()

    def calculate(self):
        self.series = pd.Series(index=[self.series.index.weekday, self.series.index.day_name()], data=self.series.values, name=self.series.name).sort_index(level=0).droplevel(0)
        
        _calculate(self)
        
        
    def __repr__(self):
        return (f'{self.__class__.__name__}(series={self.name!r}, equal_means={self.equal_means}, equal_variances={self.equal_variances})')

    def plot(self, **kwargs):
        return _datetime_plot(self, title=f'{self.name} week day distribution', **kwargs)
        
class MonthDay():  
    """
    Compare the distribution of data between month days
        arguments
        series: pandas.Series
            the Series must have a pandas.DatetimeIndex
        
        properties
        .series
            the transformed pandas.Series with month day index
        .data
            the data
        .equal_means
            the result of bluebelt.statistics.hypothesis_testing.EqualMeans().passed
        .equal_variances
            the result of bluebelt.statistics.hypothesis_testing.EqualVariances().passed

        methods
        .plot()
            plot the results as a boxplot
    """

    def __init__(self, series, **kwargs):
        
        self.series = series
        self.name = series.name
        self.calculate()

    def calculate(self):
        
        self.series = pd.Series(index=self.series.index.day, data=self.series.values, name=self.series.name).sort_index()
        
        _calculate(self)
                
    def __repr__(self):
        return (f'{self.__class__.__name__}(series={self.name!r}, equal_means={self.equal_means}, equal_variances={self.equal_variances})')

    def plot(self, **kwargs):
        return _datetime_plot(self, title=f'{self.name} month day distribution', **kwargs) 

class Week(): 
    """
    Compare the distribution of data between weeks
        arguments
        series: pandas.Series
            the Series must have a pandas.DatetimeIndex
        
        properties
        .series
            the transformed pandas.Series with week number index
        .data
            the data
        .equal_means
            the result of bluebelt.statistics.hypothesis_testing.EqualMeans().passed
        .equal_variances
            the result of bluebelt.statistics.hypothesis_testing.EqualVariances().passed

        methods
        .plot()
            plot the results as a boxplot
    """
    
    def __init__(self, series, **kwargs):
        
        self.series = series
        self.name = series.name
        self.calculate()

    def calculate(self):
        
        self.series = pd.Series(index=self.series.index.isocalendar().week, data=self.series.values, name=self.series.name).sort_index()
        
        _calculate(self)

    def __repr__(self):
        return (f'{self.__class__.__name__}(series={self.name!r}, equal_means={self.equal_means}, equal_variances={self.equal_variances})')

    def plot(self, **kwargs):
        return _datetime_plot(self, title=f'{self.name} week distribution', **kwargs)
        
class Month():
    """
    Compare the distribution of data between months
        arguments
        series: pandas.Series
            the Series must have a pandas.DatetimeIndex
        
        properties
        .series
            the transformed pandas.Series with month index
        .data
            the data
        .equal_means
            the result of bluebelt.statistics.hypothesis_testing.EqualMeans().passed
        .equal_variances
            the result of bluebelt.statistics.hypothesis_testing.EqualVariances().passed

        methods
        .plot()
            plot the results as a boxplot
    """
    
    def __init__(self, series, **kwargs):
        
        self.series = series
        self.name = series.name
        self.calculate()

    def calculate(self):
        
        self.series = pd.Series(index=self.series.index.month, data=self.series.values, name=self.series.name).sort_index()
       
        _calculate(self)
                
    def __repr__(self):
        return (f'{self.__class__.__name__}(series={self.name!r}, equal_means={self.equal_means}, equal_variances={self.equal_variances})')

    def plot(self, **kwargs):
        return _datetime_plot(self, title=f'{self.name} month distribution', **kwargs)
        
class Year():
    """
    Compare the distribution of data between years
        arguments
        series: pandas.Series
            the Series must have a pandas.DatetimeIndex
        
        properties
        .series
            the transformed pandas.Series with year index
        .data
            the data
        .equal_means
            the result of bluebelt.statistics.hypothesis_testing.EqualMeans().passed
        .equal_variances
            the result of bluebelt.statistics.hypothesis_testing.EqualVariances().passed

        methods
        .plot()
            plot the results as a boxplot
    """
    
    def __init__(self, series, **kwargs):
        
        self.series = series
        self.name = series.name
        self.calculate()

    def calculate(self):
        
        self.series = pd.Series(index=self.series.index.year, data=self.series.values, name=self.series.name).sort_index()
        
        _calculate(self)
                
    def __repr__(self):
        return (f'{self.__class__.__name__}(series={self.name!r}, equal_means={self.equal_means}, equal_variances={self.equal_variances})')

    def plot(self, **kwargs):
        return _datetime_plot(self, title=f'{self.name} year distribution')


def _calculate(dt_obj, **kwargs):

    dt_obj.frame = pd.DataFrame.from_dict({i:pd.Series(dt_obj.series[i]).to_list() for i in dt_obj.series.index.unique()}, orient='index').T    
    dt_obj.means = dt_obj.frame.mean() #{key:value for key, value in zip(dt_obj.series.index.unique(), pd.DataFrame(dt_obj.data).mean(axis=1).values)}
    dt_obj.variances = dt_obj.frame.var() #{key:value for key, value in zip(dt_obj.series.index.unique(), pd.DataFrame(dt_obj.data).var(axis=1).values)}
    dt_obj.equal_means = bluebelt.statistics.hypothesis_testing.EqualMeans(dt_obj.frame).passed
    dt_obj.equal_variances = bluebelt.statistics.hypothesis_testing.EqualVariances(dt_obj.frame).passed
    
    normal_distribution_test = bluebelt.statistics.hypothesis_testing.NormalDistribution(dt_obj.frame)

    dt_obj.normal_distribution = normal_distribution_test.passed_values
    dt_obj.normal_distribution_p_values = normal_distribution_test.p_values
        

def _datetime_plot(plot_obj, **kwargs):
        
    title = kwargs.pop('title', f'{plot_obj.name} datetime distribution')
    style = kwargs.pop('style', bluebelt.styles.paper)
    
    xlim = kwargs.pop('xlim', (None, None))
    ylim = kwargs.pop('ylim', (None, None))
    
    path = kwargs.pop('path', None)        
    
    # prepare figure
    fig, ax = plt.subplots(**kwargs)

    boxplot = ax.boxplot([plot_obj.frame[col].dropna() for col in plot_obj.frame])

    for n, box in enumerate(boxplot['boxes']):
        # add style if any is given
        box.set(**style.graphs.boxplot.boxplot.get('boxes', {}))
        
    # title
    ax.set_title(title, **style.graphs.boxplot.title)

    # labels        
    ax.set_xticklabels(plot_obj.series.index.unique())

    # set limits
    ax.set_xlim(xlim)
    ax.set_ylim(ylim)

    if path:
        plt.savefig(path)
        plt.close()
    else:
        plt.close()
        return fig


import pandas as pd
import matplotlib.pyplot as plt

import bluebelt.core.graph
import bluebelt.core.helpers



def _get_frame(_obj, **kwargs):

    columns = kwargs.get('columns', None)

    if isinstance(_obj, pd.DataFrame) and isinstance(columns, (str, list)):
        return _obj[columns]
    elif isinstance(_obj, pd.Series):
        return pd.DataFrame(_obj)
    else:
        return _obj

def _get_name(_obj, **kwargs):

    if isinstance(_obj, pd.Series):
        return _obj.name
    elif isinstance(_obj, pd.DataFrame):
        names = []
        for col in _obj.columns:
            names.append(col)
        return bluebelt.core.helpers._get_nice_list(names)
    else:
        return None
    
def line(_obj, **kwargs):
    """
    Make a line plot for a pandas Series or a pandas Dataframe
        arguments
        _obj: pandas.Series or pandas.Dataframe
        style: bluebelt style
            default value: bluebelt.styles.paper
        title: string
            default value: pandas Series name or pandas Dataframe column names
        path: string
            the full path to save the plot (e.g. 'results/plot_001.png')
            default value: None
        xlim: tuple
            a tuple with the two limits for the x-axis (e.g. (0, 100) or (None, 50))
            default value: (None, None)
        ylim: tuple
            a tuple with the two limits for the y-axis (e.g. (0, None) or (100, 200))
            default value: (None, None)
        **kwargs: all additional kwargs will be passed to matplotlib.pyplot.subplots()
    """

    style = kwargs.pop('style', bluebelt.styles.paper)
    _func = _obj._blue_function + ' ' if hasattr(_obj, '_blue_function') else ''
    title = kwargs.pop('title', f'{_get_name(_obj)} {_func}line plot')
    
    path = kwargs.pop('path', None)
    xlim = kwargs.pop('xlim', (None, None))
    ylim = kwargs.pop('ylim', (None, None))
    
    frame = _get_frame(_obj, **kwargs)

    # prepare figure
    fig, axes = plt.subplots(nrows=1, ncols=1, **kwargs)

    for col in frame:
        # _obj.plot(kind='line', ax=fig.gca())
        # bluebelt.core.graph.line(series=frame[col], ax=axes, style=style)
        axes.plot(frame[col].values, **style.graphs.line.plot, label=col)
    
    
    # xticks
    if isinstance(frame.index, pd.MultiIndex):
        bluebelt.core.helpers._get_multiindex_labels(axes, _obj, style)
        fig.subplots_adjust(bottom=.1*_obj.index.nlevels)
    elif isinstance(frame.index, pd.DatetimeIndex):
        bluebelt.core.helpers._axisformat(axes, _obj)
    
    # format things
    axes.set_xlim(xlim)
    axes.set_ylim(ylim)
    
    # legend
    if isinstance(_obj, pd.DataFrame):
        if _obj.shape[1] > 1:
            axes.legend(loc='best')
    # title
    axes.set_title(title, **style.graphs.line.title)

    if path:
        plt.savefig(path)
        plt.close()
    else:
        plt.close()
        return fig


def scatter(_obj, **kwargs):
    """
    Make a scatter plot for a pandas Series or a pandas Dataframe
        arguments
        _obj: pandas.Series or pandas.Dataframe
        style: bluebelt style
            default value: bluebelt.styles.paper
        title: string
            default value: pandas Series name or pandas Dataframe column names
        path: string
            the full path to save the plot (e.g. 'results/plot_001.png')
            default value: None
        xlim: tuple
            a tuple with the two limits for the x-axis (e.g. (0, 100) or (None, 50))
            default value: (None, None)
        ylim: tuple
            a tuple with the two limits for the y-axis (e.g. (0, None) or (100, 200))
            default value: (None, None)
        **kwargs: all additional kwargs will be passed to matplotlib.pyplot.subplots()
    """
    style = kwargs.pop('style', bluebelt.styles.paper)
    _func = _obj._blue_function + ' ' if hasattr(_obj, '_blue_function') else ''
    
    path = kwargs.pop('path', None)
    xlim = kwargs.pop('xlim', (None, None))
    ylim = kwargs.pop('ylim', (None, None))
    
    frame = _get_frame(_obj, **kwargs)

    if frame.shape[1] >= 2:
        title = kwargs.pop('title', f'{_get_name(frame.iloc[:,:2])}  {_func}scatter plot')
        index_name = frame.columns[0]
        frame = pd.DataFrame(data={frame.columns[1]: frame.iloc[:,1].values}, index=frame.iloc[:,0].values)
        frame.index.name = index_name
    else:
        title = kwargs.pop('title', f'{_get_name(_obj)}  {_func}scatter plot')
    
    
    # prepare figure
    fig, axes = plt.subplots(nrows=1, ncols=1, **kwargs)

    bluebelt.core.graph.scatter(series=frame, ax=axes, style=style)
        
    # format things
    axes.set_xlim(xlim)
    axes.set_ylim(ylim)
    
    # title
    axes.set_title(title, **style.graphs.scatter.title)

    if path:
        plt.savefig(path)
        plt.close()
    else:
        plt.close()
        return fig

def area(_obj, **kwargs):
    """
    Make an area plot for a pandas Series or a pandas Dataframe
        arguments
        _obj: pandas.Series or pandas.Dataframe
        style: bluebelt style
            default value: bluebelt.styles.paper
        title: string
            default value: pandas Series name or pandas Dataframe column names
        path: string
            the full path to save the plot (e.g. 'results/plot_001.png')
            default value: None
        xlim: tuple
            a tuple with the two limits for the x-axis (e.g. (0, 100) or (None, 50))
            default value: (None, None)
        ylim: tuple
            a tuple with the two limits for the y-axis (e.g. (0, None) or (100, 200))
            default value: (None, None)
        **kwargs: all additional kwargs will be passed to matplotlib.pyplot.subplots()
    """
    
    frame = _get_frame(_obj, **kwargs)

    style = kwargs.pop('style', bluebelt.styles.paper)
    _func = _obj._blue_function + ' ' if hasattr(_obj, '_blue_function') else ''
    title = kwargs.pop('title', f'{_get_name(_obj)} {_func}line plot')
    
    path = kwargs.pop('path', None)
    xlim = kwargs.pop('xlim', (None, None))
    ylim = kwargs.pop('ylim', (None, None))
    
    # prepare figure
    fig, axes = plt.subplots(nrows=1, ncols=1, **kwargs)

    for col in frame:
        # .get_level_values(-1) hack to get last level
        axes.stackplot(list(frame[col].index.get_level_values(-1)), frame[col].values, **style.graphs.area.stackplot)
        axes.plot(list(frame[col].index.get_level_values(-1)), frame[col].values, label=col, **style.graphs.area.plot, **kwargs)
        
    # xticks
    if isinstance(frame.index, pd.MultiIndex):
        bluebelt.core.helpers._get_multiindex_labels(axes, _obj, style)
        fig.subplots_adjust(bottom=.1*_obj.index.nlevels)
    elif isinstance(frame.index, pd.DatetimeIndex):
        bluebelt.core.helpers._axisformat(axes, _obj)
    
    # format things
    axes.set_xlim(xlim)
    axes.set_ylim(ylim)
    
    # legend
    if isinstance(_obj, pd.DataFrame):
        if _obj.shape[1] > 1:
            axes.legend(loc='best')
    
    # title
    axes.set_title(title, **style.graphs.area.title)

    if path:
        plt.savefig(path)
        plt.close()
    else:
        plt.close()
        return fig

def hist(series, **kwargs):
    """
    Make a histogram plot for a pandas Series or a pandas Dataframe
        arguments
        _obj: pandas.Series or pandas.Dataframe
        style: bluebelt style
            default value: bluebelt.styles.paper
        title: string
            default value: pandas Series name or pandas Dataframe column names
        path: string
            the full path to save the plot (e.g. 'results/plot_001.png')
            default value: None
        fit: boolean
            fit a normal distribution
            default value: False
        xlim: tuple
            a tuple with the two limits for the x-axis (e.g. (0, 100) or (None, 50))
            default value: (None, None)
        ylim: tuple
            a tuple with the two limits for the y-axis (e.g. (0, None) or (100, 200))
            default value: (None, None)
        **kwargs: all additional kwargs will be passed to matplotlib.pyplot.subplots()
    """
    
    style = kwargs.pop('style', bluebelt.styles.paper)
    _func = _obj._blue_function + ' ' if hasattr(_obj, '_blue_function') else ''
    title = kwargs.pop('title', f'{_get_name(_obj)} {_func}line plot')
    
    path = kwargs.pop('path', None)
    fit = kwargs.pop('fit', False)
    xlim = kwargs.pop('xlim', (None, None))
    ylim = kwargs.pop('ylim', (None, None))
    
    # prepare figure
    fig, axes = plt.subplots(nrows=1, ncols=1, **kwargs)

    bluebelt.core.graph.hist(series=series, ax=axes, style=style, fit=fit)
        
    # format things
    axes.set_xlim(xlim)
    axes.set_ylim(ylim)
    
    # title
    axes.set_title(title, **style.graphs.area.title)

    if path:
        plt.savefig(path)
        plt.close()
    else:
        plt.close()
        return fig

histogram = hist

def boxplot(_obj, **kwargs):
    """
    Make a boxplot for a pandas Series or a pandas Dataframe
        arguments
        _obj: pandas.Series or pandas.Dataframe
        style: bluebelt style
            default value: bluebelt.styles.paper
        title: string
            default value: pandas Series name or pandas Dataframe column names
        path: string
            the full path to save the plot (e.g. 'results/plot_001.png')
            default value: None
        xlim: tuple
            a tuple with the two limits for the x-axis (e.g. (0, 100) or (None, 50))
            default value: (None, None)
        ylim: tuple
            a tuple with the two limits for the y-axis (e.g. (0, None) or (100, 200))
            default value: (None, None)
        **kwargs: all additional kwargs will be passed to matplotlib.pyplot.subplots()
    """
    
    style = kwargs.pop('style', bluebelt.styles.paper)
    _func = _obj._blue_function + ' ' if hasattr(_obj, '_blue_function') else ''
    title = kwargs.pop('title', f'{_get_name(_obj)} {_func}line plot')
    
    path = kwargs.pop('path', None)
    xlim = kwargs.pop('xlim', (None, None))
    ylim = kwargs.pop('ylim', (None, None))
    
    frame = _get_frame(_obj, **kwargs)

    # prepare figure
    fig, axes = plt.subplots(nrows=1, ncols=1, **kwargs)

    bluebelt.core.graph.boxplot(series=frame.values, ax=axes, style=style)
        
    # format things
    axes.set_xticklabels(frame.columns)
    axes.set_xlim(xlim)
    axes.set_ylim(ylim)
            
    # title
    axes.set_title(title, **style.graphs.boxplot.title)

    if path:
        plt.savefig(path)
        plt.close()
    else:
        plt.close()
        return fig
    


def check_nan(series):

    issues = series.isna().sum()

    if issues:
        print(f'the series \'{series.name}\' contains {issues} Null value(s).')
    return issues

def check_data_types(series):

    data_types = set(series.apply(type))
    issues = int(len(data_types)>1)

    if issues:
        print(f'the series \'{series.name}\' contains {len(data_types)} data types: {data_types}')
    return issues


import numpy as np
import scipy.stats as stats


def confidence_interval_mean(series, confidence=0.95):
    n = series.size
    mean = np.mean(series)
    standard_error = series.std()/(n**0.5)
    h = standard_error * stats.t.ppf((1 + confidence) / 2, n - 1)
    return (mean - h, mean + h)

ci_mean = confidence_interval_mean
CI_mean = confidence_interval_mean

def confidence_interval_median(series, level=0.95, bootstrap=5987):
    
    # bootstrap : number of times the confidence intervals around the median should be bootstrapped (percentile method).
    # determine 95% confidence intervals of the median
    
    idx = np.random.randint(series.size, size=(5987, series.size))
    data = series.values[idx]
    medians = np.median(data, axis=1, overwrite_input=True)
    confidence_interval = np.percentile(medians, [(1-level)/0.02, (1+level)/0.02])
    return tuple(confidence_interval)

ci_median = confidence_interval_median
CI_median = confidence_interval_median

def confidence_interval_std(series, level=0.95):
    # SD*SQRT((N-1)/CHISQ.INV(1-(alpha/2), N-1))
    # SD*SQRT((N-1)/CHISQ.INV((alpha/2), N-1))
    pass


import pandas as pd
import numpy as np
import math

import matplotlib.pyplot as plt

from bluebelt.core.checks import check_kwargs

import bluebelt.core.decorators

import bluebelt.data.index


import warnings

def resolution_methods(cls):

    def sum(self):
        result = self.grouped.sum()
        result._blue_function = 'sum'
        return result

    def mean(self):
        return self.grouped.mean()

    def var(self):
        return self.grouped.var()

    def std(self):
        return self.grouped.std()

    def min(self):
        return self.grouped.min()

    def max(self):
        return self.grouped.max()
    
    def count(self):
        return self.grouped.count()
    
    def value_range(self):
        return self.grouped.apply(lambda x: x.max() - x.min())

    def inter_scale(self):
        return ( (self.grouped.sum() - self.grouped.sum().shift()) / self.grouped.sum().shift() ).abs().fillna(0)

    def intra_scale(self, **kwargs):

        values = kwargs.get('values', None)

        # _obj = self._obj[values] if isinstance(self._obj, pd.DataFrame) and values is not None else
        
        if isinstance(self._obj, pd.Series):
            index_groups = self.grouped.obj.index.names

            if (self.grouped.obj.index.names.index(self.level) + 1) < len(index_groups):
                group = index_groups[self.grouped.obj.index.names.index(self.level) + 1]
                result = self.grouped.apply(lambda s: s.groupby(group).sum()).unstack(level=-1)
                result = pd.Series((result - result.shift().multiply((result.sum(axis=1) / result.sum(axis=1).shift()), axis=0)).abs().sum(axis=1) / (result.sum(axis=1) * 2), name=self._obj.name)
            else:
                result = pd.Series(index=self.grouped.sum().index, data=[0]*self.grouped.sum().size, name=self._obj.name)

            return result
        else:
            raise ValueError("intra_scale can not be called form a pandas.DataFrame Groupby object")
    

    def subsize_count(self, count=3, size=1):
        """
        Count the number of times a list of <count> items with size <size> fit in the groupby object (which is a pandas Series)
        e.g.
        groupby object: pd.Series([10, 8, 3, 3, 5])
        count = 3
        size = 1

        returns 9

        step 0: (3, 3, 5, 8, 10)
        step 1: (3, 3, 4, 7, 9)
        step 2: (3, 3, 3, 6, 8)
        step 3: (2, 3, 3, 5, 7)
        step 4: (2, 2, 3, 4, 6)
        step 5: (2, 2, 2, 3, 5)
        step 6: (1, 2, 2, 2, 4)
        step 7: (1, 1, 1, 2, 3)
        step 8: (0, 1, 1, 1, 2)
        step 9: (0, 0, 0, 1, 1)

        """
        if isinstance(count, (float, int)):
            count = [int(count)]
        
        result = {}
        for c in count:
            result[c] = self.grouped.apply(lambda x: _subsize_count(series=x, count=c, size=size)).values
        self.grouped = pd.DataFrame(result, index=self.grouped.groups.keys()) #self.grouped.apply(lambda x: _subsize_count(series=x, count=count, size=size))
        self.func = f'count subsize (count={count}, size={size})'
        
        if len(count) == 2:
            d = {}
            #arr = self.grouped.values.T
            for val in range(int(self.grouped.values.min()), int(self.grouped.values.max())):
                
                if self.grouped.iloc[:,0].mean() > self.grouped.iloc[:,1].mean():
                #if arr[0].mean() > arr[1].mean():
                    d[val] = self.grouped.shape[0] - self.grouped[(self.grouped.iloc[:,0]>val) & (self.grouped.iloc[:,1]<val)].shape[0]
                    #d[val] = arr[0][arr[0]<val].size + arr[1][arr[1]>val].size
                else:
                    d[val] = self.grouped.shape[0] - self.grouped[(self.grouped.iloc[:,0]<val) & (self.grouped.iloc[:,1]>val)].shape[0]
                    #d[val] = arr[0][arr[0]>val].size + arr[1][arr[1]>val].size

            out_of_bounds = np.min(list(d.values()))
            keys = list({key:value for (key,value) in d.items() if value == out_of_bounds}.keys())
            
            self.optimum = (np.min(keys), np.max(keys))
            self.out_of_bounds = out_of_bounds
        else:
            self.optimum = None
            self.out_of_bounds = None

        return self

    def subseries_count(self, subseries=None, **kwargs):
        """
        Count the number of times a subseries of wil fit in the groupby object (which is a pandas Series)
        
        Under construction!! Not valid on extreme value differences...

        e.g.
        groupby object: pd.Series([10, 8, 3, 3, 5])
        subseries: pd.Series([1, 2, 3])

        returns 4

        step 0: (3, 3, 5, 8, 10)
        step 1: (3, 3, 4, 6, 7)
        step 2: (3, 3, 3, 4, 4)
        step 3: (2, 3, 3, 2, 1)
        step 4: (1, 1, 0, 2, 1)

        """
        self.grouped = self.grouped.apply(lambda x: _subseries_count(series=x, subseries=subseries, **kwargs))
        self.func = f'count subseries (subseries={list(subseries)})'
        return self

    def graph(self, **kwargs):

        if self.func == '':
            series = self.grouped.sum()
            func = 'sum'
        else:
            series = self.grouped
            func = self.func

        style = kwargs.pop('style', bluebelt.styles.paper)
        path = kwargs.pop('path', None)

        xlim = kwargs.pop('xlim', (None, None))
        ylim = kwargs.pop('ylim', (None, None))

        fig, ax = plt.subplots(**kwargs)

        # observations
        ax.plot(series, **style.resolution.plot)

        xlims = ax.get_xlim()
        
        # mean
        if hasattr(self, 'optimum'):
            if self.optimum is not None:
                ax.fill_between(xlims, self.optimum[0], self.optimum[1], **style.resolution.optimum_fill_between)
                ax_text = f'bounds: {self.optimum[0]:1.0f} - {self.optimum[1]:1.0f}\nout of bounds: {self.out_of_bounds}'
                ax.text(0.02, 0.98, ax_text, transform=ax.transAxes, **style.resolution.bounds_text)
            else:
                ax.axhline(series.values.mean(), **style.resolution.axhline)
                ax.text(series.index.values.min(), series.values.mean(), f'{series.values.mean():1.2f}', **style.resolution.text)

        # labels
        ax.set_title(f'{self._obj.name} grouped', **style.resolution.title)
        ax.set_xlabel(self.level)
        ax.set_ylabel(func)

        #reset xlim
        ax.set_xlim(xlims)

        ax.set_xlim(xlim)
        ax.set_ylim(ylim)
    

        if path:
            plt.savefig(path)
            plt.close()
        else:
            plt.close()
            return fig

    setattr(cls, 'sum', sum)
    setattr(cls, 'mean', mean)
    setattr(cls, 'var', var)
    setattr(cls, 'std', std)
    setattr(cls, 'min', min)
    setattr(cls, 'max', max)
    setattr(cls, 'count', count)
    setattr(cls, 'value_range', value_range)
    setattr(cls, 'intra_scale', intra_scale)
    setattr(cls, 'inter_scale', inter_scale)
    setattr(cls, 'subsize_count', subsize_count)
    setattr(cls, 'subseries_count', subseries_count)
    
    return cls

@bluebelt.core.decorators.class_methods
@resolution_methods
class GroupByDatetimeIndex():
    """
    Group a pandas.Series or pandas.DataFrame by DateTime index and apply a specific function.
        arguments
        series: pandas.Series
        how: str
            a string with date-time keywords that can be parsed to group the index
            keywords:
            
            default value "week"

        Apply one of the following functions:
            .sum()
            .mean()
            .min()
            .max()
            .std()
            .value_range()
            .count()
            .subsize_count()

        e.g. series.blue.data.group_index(how="week").sum()
        
    """
    
    def __init__(self, _obj, level="week", iso=True, **kwargs):

        self._obj = _obj
        self.level = level
        self.iso = iso
        self.nrows = self._obj.shape[0]
        self.calculate(**kwargs)

    def calculate(self, **kwargs):

        if self.iso and self.level in ['month']:
            warnings.warn(f'rule={self.level} is not possible when using iso=True; iso will be set to False', Warning)
            self.iso = False
        elif not self.iso and self.level in ['week']:
            warnings.warn(f'rule={self.level} is not possible when using iso=False; iso will be set to True', Warning)
            self.iso = True
        
        if self.iso:
            self._obj = bluebelt.data.index.iso_index(self._obj)
        else:
            self._obj = bluebelt.data.index.dt_index(self._obj)

        self.grouped = self._obj.groupby(self._obj.index.names[:self._obj.index.names.index(self.level) + 1])

    def __str__(self):
        return ""
    
    def __repr__(self):
        return self.grouped.__repr__()

def _subsize_count(series, count=3, size=1):
    series = pd.Series(series)/size
    result = series.sum()*count
    for i in range(count, 0, -1):
        result = min(result, math.floor(series.nsmallest(len(series) - count + i).sum() / i))
    return result

def _subseries_count(series, subseries=None, **kwargs):
    series = pd.Series(series)
    subseries = pd.Series(subseries)
    result=series.sum()*subseries.sum()
    for i in range(len(subseries), 0, -1):
        result = min(result, math.floor(series.nsmallest(len(series) - len(subseries) + i).sum() / subseries.nsmallest(i).sum()))
    return result


def _get_index_dict(self):

    if self.iso:
        _dict = {
            'year': self._obj.index.isocalendar().year.rename('year'), # year including the century
            'week': self._obj.index.isocalendar().week.rename('week'), # iso-week of the year (1 to 53)
            'day': self._obj.index.isocalendar().day.rename('day'), # day of the week (1 to 7)
        }
    else:
        _dict = {
            'year': self._obj.index.year.rename('year'), # year including the century
            'month': self._obj.index.month.rename('month'), # month (1 to 12)
            'day': self._obj.index.day.rename('day'), # day of the month (1 to 31)
            'hour': self._obj.index.hour.rename('hour'), # hour, using a 24-hour clock (0 to 23)
            'minute': self._obj.index.minute.rename('minute'),
            'second': self._obj.index.second.rename('second'),
        }

    return _dict


def _get_index(self):
    index = []
    
    level = min(len(self._obj.index.names), self._obj.index.names.index(self.level) + 1)

    for group, values in enumerate(self._obj.index.levels[:level]):
        if values.size > 1:
            index.append(self._obj.index.names[group]) 

    return index


import datetime
import pandas as pd
import numpy as np

# Dutch holidays

def holidays(check_date, **kwargs):

    holidays = kwargs.get('holidays', {
        'new years day': True,
        'good friday': True,
        'easter': True,
        'easter monday': True,
        'ascension day': True,
        'pentecost': True,
        'whit monday': True,
        'christmas': True,
        'boxing day': True,
        'kings day': True,
        'liberation day': True,
        })

    #deal with datetimes
    check_date = pd.to_datetime(check_date).date()

    holiday_dates={'new years day': _new_years_day(check_date.year),
                   'good friday': _good_friday(check_date.year),
                   'easter': _easter(check_date.year),
                   'easter monday': _easter_monday(check_date.year),
                   'ascension day': _ascension_day(check_date.year),
                   'pentecost': _pentecost(check_date.year),
                   'whit monday': _whit_monday(check_date.year),
                   'christmas': _christmas(check_date.year),
                   'boxing day': _boxing_day(check_date.year),
                   'kings day': _kings_day(check_date.year),
                   'liberation day': _liberation_day(check_date.year)}

    for key, value in holidays.items():
        if value==True:
            if holiday_dates.get(key)==check_date:
                return True
    return False

def _new_years_day(year):
    return datetime.date(year, 1, 1)

def _easter(year):
    "Returns Easter as a date object."
    a = year % 19
    b = year // 100
    c = year % 100
    d = (19 * a + b - b // 4 - ((b - (b + 8) // 25 + 1) // 3) + 15) % 30
    e = (32 + 2 * (b % 4) + 2 * (c // 4) - d - (c % 4)) % 7
    f = d + e - 7 * ((a + 11 * d + 22 * e) // 451) + 114
    month = f // 31
    day = f % 31 + 1
    return datetime.date(year, month, day)

def _easter_monday(year):
    return _easter(year) + datetime.timedelta(days=1)

def _good_friday(year):
    return _easter(year) + datetime.timedelta(days=-2)

def _ascension_day(year):
    return _easter(year) + datetime.timedelta(days=39)

def _pentecost(year):
    return _easter(year) + datetime.timedelta(days=49)

def _whit_monday(year):
    return _easter(year) + datetime.timedelta(days=50)

def _christmas(year):
    return datetime.date(year, 12, 25)

def _boxing_day(year):
    return datetime.date(year, 12, 26)

def _kings_day(year):
    return datetime.date(year, 4, 27)

def _liberation_day(year, annually=False):
    if annually==True or year % 5 == 0:
        return datetime.date(year, 5, 5)
    else:
        return np.datetime64("NaT")



import bluebelt.styles.defaults as defaults

# main plot | pyplot.plot
scatter = {
    'linewidth': 0,
    'marker': 'o',
    'zorder': 50,
    'color': defaults.black+(1,),
}

plot = {
    'marker': None,
    'color': defaults.red,
    'linewidth': 1,
    'linestyle': 'solid',
    'zorder': 30, 
}

# stat text | pyplot.text
stat_text = {
    'ha': 'right',
    'va': 'bottom',
    'zorder': 90,
}

# array text | pyplot.text
array_text = {
    'ha': 'left',
    'va': 'top',
    'zorder': 90,
}

# plot title | pyplot.set_title
title = {
    'loc': 'left',
    'size': defaults.small
}

# plot title | fig.suptitle
suptitle = {
    'ha': 'left',
    'size': defaults.medium
}



import bluebelt.styles.paper.graphs.line
import bluebelt.styles.paper.graphs.scatter
import bluebelt.styles.paper.graphs.area
import bluebelt.styles.paper.graphs.hist
import bluebelt.styles.paper.graphs.boxplot



import pandas as pd
import numpy as np

import bluebelt.core.decorators

@bluebelt.core.decorators.class_methods
class Subset():
    """
    Create a subset based on the pandas.Dataframe column values        
        arguments
        inverse: boolean
            if True then the inverse of the filtered pandas Dataframe will be returned
            default value: False
        
        usage
        df
            | column_1  | column_2  | column_3  |
        -----------------------------------------
        0   | a         | a         | a         |
        1   | b         | a         | b         |
        2   | a         | b         | c         |
        3   | b         | b         | a         |
        4   | a         | c         | b         |
        5   | b         | c         | c         |
        6   | a         | a         | a         |
        7   | b         | a         | b         |
        8   | a         | b         | c         |
        9   | b         | b         | a         |
        

        subset = bluebelt.data.subsets.Subset(frame, column_1="a", column_2=["a", "b"])
        subset
            | column_1  | column_2  | column_3  |
        -----------------------------------------
        0   | a         | a         | a         |
        2   | a         | b         | c         |
        6   | a         | a         | a         |
        8   | a         | b         | c         |
        
        
    """
    def __init__(self, frame, inverse=False, **kwargs):

        self.frame = frame
        self.inverse = inverse
        self.kwargs = kwargs
        self.calculate(**kwargs)

    def calculate(self, **kwargs):

        # build filters
        filters={}
        for col in self.kwargs:
            if col in self.frame.columns:
                values = self.kwargs.get(col) if isinstance(self.kwargs.get(col), list) else [self.kwargs.get(col)]
                for value in values:
                    if value not in self.frame[col].values:
                        raise ValueError(f'{value} is not in {col}')
                filters[col]=values
            else:
                raise ValueError(f'{col} is not in frame')

        self.filters = filters

        # filter the frame
        if self.inverse:
            frame=self.frame[self.frame.isin(filters).sum(axis=1) != len(filters.keys())]
        else:
            frame=self.frame[self.frame.isin(filters).sum(axis=1) == len(filters.keys())]
            
        self.result = frame

    def __str__(self):
        return ""
    
    def __repr__(self):
        return (f'{self.__class__.__name__}(frame_length={self.frame.shape[0]:1.0f}, result_length={self.result.shape[0]:1.0f}, filter={self.filters})')



import pandas as pd
import numpy as np
import scipy.stats as stats

import warnings

import matplotlib.pyplot as plt

import bluebelt.core.helpers
import bluebelt.core.decorators

import bluebelt.styles

@bluebelt.core.decorators.class_methods
class Polynomial():
    """
    Find the polynomial of a series and project a bandwidth

    Polynomial(series, shape=(0, 6), validation='p_value', threshold=0.05, confidence=0.8, outlier_sigma=2, adjust=True, **kwargs)
    
    series: pandas.Series
    shape: int or tuple
        when an int is provided the polynomial is provided as n-th degree polynomial
        when a tuple is provided the function will find an optimised polynomial between first and second value of the tuple
        default value: (0, 6)
    validation: string
        validation type for shape tuple
        p_val: test for normal distribution of the residuals
        rsq: check for improvement of the rsq value
        default value: p_val
    threshold: float
        the threshold for normal distribution test or rsq improvement
        default value: 0.05
    confidence: float
        the bound confidence
        default value: 0.8
    outlier_sigma: float
        outliers are datapoints outside the outlier_sigma fraction
        default value: 2
    adjust: boolean
        adjust polynomial for outliers
        default value: True
    """
    def __init__(self, series, shape=(0, 6), validation='p_value', threshold=0.05, confidence=0.8, outlier_sigma=2, adjust=True, **kwargs):
        
        self.series = series
        self.shape = shape
        self.validation = validation
        self.threshold = threshold
        self.confidence = confidence
        self.outlier_sigma = outlier_sigma
        self.adjust = adjust

        self.calculate()

    def calculate(self):
        
        # set pattern and residuals
        self.result, self.residuals, self.shape, self.polyfit, self.statistic, self.p_value, self.rsq = _poly_hand_granade(series=self.series, shape=self.shape, validation=self.validation, threshold=self.threshold)

        # set outliers
        self.outliers = pd.Series(data=np.where(self.residuals.abs() > self.residuals.std() * self.outlier_sigma, self.series, None), index=self.series.index)
        self.outliers_count = np.count_nonzero(self.outliers)

        # handle adjusted
        self.adjusted = self.series.loc[~self.outliers.notnull()]
        if self.adjust:
            # replace outliers with None values so they will be ignored by _poly_hand_granade and reset pattern
            values = pd.Series(data=np.where(self.outliers.notnull(), None, self.series).astype(np.float), index=self.series.index)
            self.result, self.residuals, self.shape, self.polyfit, self.statistic, self.p_value, self.rsq = _poly_hand_granade(series=values, shape=self.shape, validation=self.validation, threshold=self.threshold)
        
        # handle bounds
        _calculate_bounds(self)

    def __repr__(self):
        return (f'{self.__class__.__name__}(n={self.series.size:1.0f}, shape={self.shape}, validation={self.validation}, threshold={self.threshold}, confidence={self.confidence}, outlier_sigma={self.outlier_sigma}, adjust={self.adjust}, outliers={self.outliers_count})')
    
    def plot(self, **kwargs):
        
        return _pattern_plot(self, **kwargs)

@bluebelt.core.decorators.class_methods
class Periodical():
    
    """
    Find the periodical pattern of a series and project a bandwidth

    Periodical(series, rule='1W', how='mean', resolution=None, confidence=0.8, outlier_sigma=2, adjust=True, **kwargs)
    
    series: pandas.Series
    rule: period representation used for resampling the series
        default value: "1W"
    how: define how the period must be evaluated
        options are "mean", "min", "max" and "std"
        default value: "mean"
    resolution: define the resolution of the pattern
        the pattern is rounded to fit the resolution
        default value: None
    confidence: float
        the bandwidth confidence
        default value: 0.8
    outlier_sigma: float
        outliers are datapoints outside the outlier_sigma fraction
        default value: 2
    
    """

    def __init__(self, series, rule='1W', how='mean', resolution=None, confidence=0.8, outlier_sigma=2, adjust=True, **kwargs):
        
        self.series = series
        self.rule = rule
        self.how = how
        self.resolution = resolution
        self.confidence = confidence
        self.outlier_sigma = outlier_sigma
        self.adjust = adjust
        
        self.calculate()

    def calculate(self):
        
        # set pattern and residuals        
        self.result, self.residuals, self.statistic, self.p_value, self.rsq = _peri_hand_granade(series=self.series, rule=self.rule, how=self.how, resolution=self.resolution)

        # set outliers
        self.outliers = pd.Series(data=np.where(self.residuals.abs() > self.residuals.std() * self.outlier_sigma, self.series, None), index=self.series.index)
        self.outliers_count = np.count_nonzero(self.outliers)

        # handle adjusted
        self.adjusted = self.series.loc[~self.outliers.notnull()]
        if self.adjust:
            # replace outliers with None values so they will be ignored by _peri_hand_granade and reset pattern
            values = pd.Series(data=np.where(self.outliers.notnull(), None, self.series).astype(np.float), index=self.series.index)
            self.result, self.residuals, self.statistic, self.p_value, self.rsq = _peri_hand_granade(series=values, rule=self.rule, how=self.how, resolution=self.resolution)
        
        self.result.rename(f'periodical ({self.rule})', inplace=True)

        # handle bounds
        _calculate_bounds(self)

    def __repr__(self):
        return (f'{self.__class__.__name__}(n={self.series.size:1.0f}, rule={self.rule}, how={self.how}, resolution={self.resolution}, confidence={self.confidence}, outlier_sigma={self.outlier_sigma}, adjust={self.adjust}, outliers={self.outliers_count})')
    
    def plot(self, **kwargs):
        return _pattern_plot(self, **kwargs)

# helper functions
def _poly_hand_granade(series, shape=(0, 6), validation='p_value', threshold=0.05, **kwargs):

    """
    Find the polynomial of a series.
    series = the pandas Series
    shape: int or tuple
        when an int is provided the polynomial is provided as n-th degree polynomial
        when a tuple is provided the function will find an optimised polynomial between first and second value of the tuple
    validation: only for tuple shape
        p_value: test for normal distribution of the residuals
        rsq: check for improvement of the rsq value
    threshold: the threshold for normal distribution test or rsq improvement
    """

    try:
        # get the index
        index = series.index.view(float)-series.index.view(float).min()
        index = index / np.gcd.reduce(index)

        # drop nan values
        _index = series.dropna().index.view(float)-series.index.view(float).min()
        _index = _index / np.gcd.reduce(_index)
    except:
        # the index is not convertable
        index = np.arange(0, series.index.size)
        _index = pd.Series(index=np.arange(0,series.size), data=series.values).dropna().index

    # get the values
    values = series.dropna().values

    # set first rsq
    _rsq = 0


    if isinstance(shape, int):
        polyfit = np.polynomial.polynomial.polyfit(_index, values, shape)
        pattern = pd.Series(index=series.index, data=np.polynomial.polynomial.polyval(index, polyfit), name=f'{_get_nice_polynomial_name(shape)}')
        residuals = pd.Series(index=series.index, data=[a - b for a, b in zip(series.values, pattern)])
        
        statistic, p_value = stats.normaltest(residuals.dropna().values)
        rsq = np.corrcoef(series.values, pattern.values)[1,0]**2

    elif isinstance(shape, tuple):

        with warnings.catch_warnings():
            warnings.filterwarnings('error')
            for shape in range(shape[0], shape[1]+1):
                try:
                    polyfit = np.polynomial.polynomial.polyfit(_index, values, shape)
                    pattern = pd.Series(index=series.index, data=np.polynomial.polynomial.polyval(index, polyfit), name=f'{_get_nice_polynomial_name(shape)}')
                    residuals = pd.Series(index=series.index, data=[a - b for a, b in zip(series.values, pattern)])
                    
                    np_err = np.seterr(divide='ignore', invalid='ignore') # ignore possible divide by zero
                    rsq = np.corrcoef(series.values, pattern.values)[1,0]**2
                    np.seterr(**np_err) # go back to previous settings
                    
                    statistic, p_value = stats.normaltest(residuals.dropna().values)

                    if validation=='p_value' and p_value >= threshold:
                        break
                    elif validation=='rsq' and (rsq - _rsq) / rsq < threshold:
                        shape = shape - 1
                        polyfit = np.polynomial.polynomial.polyfit(_index, values, shape)
                        pattern = pd.Series(index=series.index, data=np.polynomial.polynomial.polyval(index, polyfit), name=f'{_get_nice_polynomial_name(shape)}')    
                        residuals = pd.Series(index=series.index, data=[a - b for a, b in zip(series.values, pattern)])
                        
                        # reset rsq
                        rsq = _rsq
                        break
                    
                    # set previous rsq to current rsq
                    _rsq = rsq

                except np.polynomial.polynomial.pu.RankWarning:
                    print(f'RankWarning at {_get_nice_polynomial_name(shape)}')
                    break
    else:
        pattern = None
        residuals = None

    return pattern, residuals, shape, polyfit, statistic, p_value, rsq

def _peri_hand_granade(series, rule, how, resolution, **kwargs):

    # set pattern and residuals
    if how=='mean':
        pattern = series.resample(rule=rule).mean()
    elif how=='min':
        pattern = series.resample(rule=rule).min()
    elif how=='max':
        pattern = series.resample(rule=rule).max()
    elif how=='std':
        pattern = series.resample(rule=rule).std()
    else:
        pattern = series.resample(rule=rule).sum()
    
    # reindex pattern
    if any([period for period in ['M', 'A', 'Q', 'BM', 'BA', 'BQ', 'W'] if (period == "".join(char for char in rule if not char.isnumeric()))]):
        pattern = pattern.reindex_like(series, method = 'bfill')
    else:
        pattern = pattern.reindex_like(series, method = 'ffill')

    if resolution:
        # adjust for resolution
        pattern = pattern.divide(resolution).round(0).multiply(resolution)
    
    residuals = series - pattern

    statistic, p_value = stats.normaltest(residuals.dropna().values)
    rsq = np.corrcoef(series.values, pattern.values)[1,0]**2

    return pattern, residuals, statistic, p_value, rsq

def _get_nice_polynomial_name(shape):
    if shape==0:
        return 'linear'
    if shape==1:
        return str(shape)+'st degree polynomial'
    elif shape==2:
        return str(shape)+'nd degree polynomial'
    elif shape==3:
        return str(shape)+'rd degree polynomial'
    else:
        return str(shape)+'th degree polynomial'

def _calculate_bounds(_obj):
        
    _obj.sigma_level = stats.norm.ppf(1-(1-_obj.confidence)/2)

    # set bounds
    _obj.upper = _obj.result + _obj.residuals.std() * _obj.sigma_level
    _obj.lower = _obj.result - _obj.residuals.std() * _obj.sigma_level
    _obj.bounds = _obj.residuals.std() * _obj.sigma_level

    # set out of bounds values
    _obj.out_of_bounds = _obj.series[((_obj.series > _obj.upper) | (_obj.series < _obj.lower)) & (_obj.outliers.isnull())]

    return _obj

def _pattern_plot(_obj, **kwargs):
        
    style = kwargs.pop('style', bluebelt.styles.paper)
    path = kwargs.pop('path', None)

    bounds = kwargs.pop('bounds', True)
    residuals = kwargs.pop('residuals', True)
    
    xlim = kwargs.pop('xlim', (None, None))
    ylim = kwargs.pop('ylim', (None, None))
    
    # prepare figure
    fig = plt.figure(constrained_layout=False, **kwargs)
    if residuals:
        gridspec = fig.add_gridspec(nrows=2, ncols=1, height_ratios=[5,3], wspace=0, hspace=0)
        ax2 = fig.add_subplot(gridspec[1, 0], zorder=40)
    else:
        gridspec = fig.add_gridspec(nrows=1, ncols=1)
        
    ax1 = fig.add_subplot(gridspec[0, 0], zorder=50)
    
    # observations
    ax1.plot(_obj.series, **style.patterns.observations)

    # pattern
    ax1.plot(_obj.result, **style.patterns.pattern)
    
    # outliers
    ax1.plot(_obj.outliers, **style.patterns.outlier_background)
    ax1.plot(_obj.outliers, **style.patterns.outlier)

    # bounds
    if bounds:
        ax1.fill_between(_obj.series.index, _obj.lower, _obj.upper, label=f'{(_obj.confidence * 100):1.0f}% bounds', **style.patterns.bandwidth_fill_between)
        ax1.plot(_obj.lower, **style.patterns.lower)
        ax1.plot(_obj.upper, **style.patterns.upper)
    
        # out of bounds
        ax1.plot(_obj.out_of_bounds, **style.patterns.out_of_bounds)

        ax1.legend(loc='best')
        
    # labels
    ax1.set_title(f'{_obj.series.name} {_obj.result.name}', **style.patterns.title)
    ax1.set_ylabel('value')

    # set x axis locator
    bluebelt.core.helpers._axisformat(ax1, _obj.series)

    if residuals:
        # residuals histogram
        ax2.hist(_obj.residuals, **style.patterns.histogram)
        ax2.set_yticks([])

        # get current limits
        xlims = ax2.get_xlim()
        ylims = ax2.get_ylim()
        
        # fit a normal distribution to the data
        norm_mu, norm_std = stats.norm.fit(_obj.residuals.dropna())
        pdf_x = np.linspace(xlims[0], xlims[1], 100)
        ax2.plot(pdf_x, stats.norm.pdf(pdf_x, norm_mu, norm_std), **style.patterns.normal_plot)

        # histogram x label
        ax2.set_xlabel('residuals distribution')
                
        ax2.set_ylim(ylims[0], ylims[1]*1.5)
        ax2.spines['left'].set_visible(False)
        ax2.spines['right'].set_visible(False)

        ax2.text(0.02, 0.7, f'D’Agostino-Pearson\nstatistic: {_obj.statistic:1.2f}\np: {_obj.p_value:1.2f}', transform=ax2.transAxes, **style.patterns.statistics)

    ax1.set_xlim(xlim)
    ax1.set_ylim(ylim)
    
    if path:
        plt.savefig(path)
        plt.close()
    else:
        plt.close()
        return fig


import bluebelt.styles.defaults as defaults


hist = {
    'density': True, 
    'alpha': 1, # alpha fraction 0-1
    'edgecolor': None, # color name, hex or None
    'facecolor': defaults.blue, # color name, hex or None
    'fill': False, #True, False
    'hatch': '//', # '/', '\', '|', '-', '+', 'x', 'o', 'O', '.' or '*'
    'linestyle': 'solid', # '-', '--', '-.', ':', '', (offset, on-off-seq), ...
    'linewidth': 1, # float
}

# plot used for normal distribution | pyplot.plot
normal_plot = {
    'color': defaults.blue,
    'linestyle': 'dashed',
    'linewidth': 1,
}


import pandas as pd
import math

def class_methods(cls):

    @property
    def attributes(self):
        for attr, value in self.__dict__.items():
            if isinstance(value, (str, int, float, list, tuple)):
                print(f'{attr}: {value}')
            else:
                print(f'{attr}: {type(value)}')

    @property
    def help(self):
        print(self.__doc__)

    setattr(cls, 'attributes', attributes)
    setattr(cls, 'help', help)
    
    return cls

def add_docstring(value):
    if not isinstance(value, str):
        value = value.__doc__
    def _doc(func):
        func.__doc__ = value
        return func
    return _doc

# resampling
resolution_functions = {
    'sum': sum,
    'min': min,
    'max': max,
    }

def resolution_function(func):
    resolution_functions[func.__name__] = func
    return func



import bluebelt.styles.defaults as defaults

# main plot | pyplot.plot
observations = {
    'linewidth': 0,
    'marker': 'o',
    'zorder': 50,
}

# pattern plot | pyplot.plot
pattern = {
    'marker': None,
    'linewidth': 1,
    'zorder': 99,
    'color': defaults.blue+(1,),
    'linestyle': '-',
}

# out of bounds values | pyplot.plot
out_of_bounds = {
    'linewidth': 0,
    'marker': 'o',
    'markerfacecolor': defaults.blue+(1,),
    #'markeredgecolor': defaults.red+(1,),
    'zorder': 55,
}

# outlier background | pyplot.plot
outlier_background = {
    'linewidth': 0,
    'marker': 'o',
    'markersize': 17,
    'markerfacecolor': defaults.white,
    'zorder': 40, 
}

# outlier marker | pyplot.plot
outlier = {
    'linewidth': 0,
    'marker': 'o',
    'markersize': 9,
    'markerfacecolor': defaults.white,
    'markeredgecolor': defaults.black+(1,),
    'markeredgewidth': 1,
    'zorder': 45, 
}



# bandwidth fill_between | pyplot.fill_between
bandwidth_fill_between = {
    'color': None,
    'hatch': '////',
    'edgecolor': defaults.blue+(1,), # color name, hex or None
    'linestyle': '-',
    'linewidth': 0,
}

# lower plot | pyplot.plot
lower = {
    'marker': None,
    'linewidth': 1,
    'zorder': 99,
    'color': defaults.blue+(1,),
    'linestyle': '--',
}

# upper plot | pyplot.plot
upper = {
    'marker': None,
    'linewidth': 1,
    'zorder': 99,
    'color': defaults.blue+(1,),
    'linestyle': '--',
}

# histogram | pyplot.hist
histogram = {
    'density': True, 
    'alpha': 1, # alpha fraction 0-1
    'edgecolor': None, # color name, hex or None
    'facecolor': defaults.blue, # color name, hex or None
    'fill': False, #True, False
    'hatch': '//', # '/', '\', '|', '-', '+', 'x', 'o', 'O', '.' or '*'
    'linestyle': 'solid', # '-', '--', '-.', ':', '', (offset, on-off-seq), ...
    'linewidth': 1, # float
}

# stat text | pyplot.text
statistics = {
    'ha': 'left',
    'va': 'top',
    'zorder': 90,
    'size': defaults.xsmall,
}
# plot used for normal distribution | pyplot.plot
normal_plot = {
    'color': defaults.blue,
    'linestyle': 'dashed',
    'linewidth': 1,
}

# plot title | pyplot.set_title
title = {
    'loc': 'left',
}


import pandas as pd
import numpy as np
import scipy.stats as stats
import matplotlib.pyplot as plt
from matplotlib.ticker import FormatStrFormatter

# import bluebelt.analysis.ci as ci
import bluebelt.core.helpers
import bluebelt.styles


class Correlation():
        
    def __init__(self, frame, columns=None, confidence=0.95, **kwargs):
        
        # check arguments
        if not isinstance(frame, pd.DataFrame):
            raise ValueError('frame is not a Pandas DataFrame')

        if columns is not None and not isinstance(columns, list):
            raise ValueError('columns is not a list')
        
        self.frame = frame if columns is None else frame[columns]
        self.columns = columns if columns is not None else self.frame.columns.to_list()
        self.nrows = self.frame.shape[0]
        self.ncols = self.frame.shape[1]
        self.confidence = confidence
            
        self.calculate()

    def calculate(self):
        self.z = stats.norm.ppf(1-(1-self.confidence)/2)
        self.rci = self._r_ci()

    def __str__(self):
        precision = 3

        result =f'Pearson correlation coefficient with {self.confidence*100:1.1f}% confidence interval\n\n'
        result += self._r_ci().applymap(lambda x: np.round(x,precision) if isinstance(x, (np.floating, int, float, tuple)) else x).to_string()
        
        return result
    
    def __repr__(self):
        return (f'{self.__class__.__name__}(columns={self.columns}, nrows={self.nrows}, ncols={self.ncols}, confidence={self.confidence})')
    
    @property
    def result(self, precision=3):
        print(self._r_ci().applymap(lambda x: np.round(x,precision) if isinstance(x, (np.floating, int, float, tuple)) else x).to_string())
    
    def plot(self, **kwargs):
        
        style = kwargs.get('style', bluebelt.styles.paper)
        kwargs.pop('style', None)
        path = kwargs.pop('path', None)

        fig, ax = plt.subplots(nrows=self.ncols, ncols=self.ncols, **kwargs)
        for row, row_name in enumerate(self.columns):
            for col, col_name in enumerate(self.columns):
                if row==col:
                    # plot a histogram where row and column are equal
                    ax[row, col].hist(self.frame[row_name], **style.correlation.histogram)
                else:
                    # plot a scatter plot
                    ax[row, col].plot(self.frame[col_name], self.frame[row_name], **style.correlation.scatter)
                    
                    # fetch r and ci
                    r_value = self.rci.iloc[row, col*2]
                    rho_l = self.rci.iloc[row, 1+col*2][0]
                    rho_u = self.rci.iloc[row, 1+col*2][1]
                    ax[row, col].text(0.01, 0.99, f'r :{r_value:1.2f} CI: ({rho_l:1.2f} - {rho_u:1.2f})', transform=ax[row, col].transAxes, **style.correlation.text)
                    
                    # share X-axis
                    if col==0 and row>1:
                        ax[row, col].sharex(ax[1, col])
                    elif col>0 and row>0:
                        ax[row, col].sharex(ax[0, col])
                        
                    # share Y-axis
                    if row==0 and col>1:
                        ax[row, col].sharey(ax[row, 1])
                    elif row>0 and col>0:
                        ax[row, col].sharey(ax[row, 0])
                                        
                        
                if row==self.ncols-1:
                    ax[row, col].set_xlabel(col_name)
                
                
                if col==0:
                    ax[row, col].set_ylabel(row_name)
                
                
                ax[row, col].set_xticks([])
                ax[row, col].set_yticks([])
                
        plt.subplots_adjust(hspace = 0)
        plt.subplots_adjust(wspace = 0)
        ax[0, 0].set_title(f'correlation plot for {bluebelt.core.helpers._get_nice_list(self.columns)}', **style.correlation.title)
    
        plt.close()
        return fig

    def _r_ci(self):
        
        cols = []
        for col in range(0, self.ncols):
            rows = []
            for row in range(0, self.ncols):

                # calculate r and ci
                if row != col:
                    r_value = (np.corrcoef(self.frame[self.columns[row]], self.frame[self.columns[col]])[0,1])
                        
                    z_l = 0.5 * (np.log(1+r_value) - np.log(1-r_value)) - self.z / (self.frame[[self.columns[row], self.columns[col]]].dropna(axis=0).shape[0]-3)
                    z_u = 0.5 * (np.log(1+r_value) - np.log(1-r_value)) + self.z / (self.frame[[self.columns[row], self.columns[col]]].dropna(axis=0).shape[0]-3)
                    rho_l = (np.exp(2 * z_l) - 1) / (np.exp(2 * z_l) + 1)
                    rho_u = (np.exp(2 * z_u) - 1) / (np.exp(2 * z_u) + 1)

                    ci = [r_value, (rho_l, rho_u)]
                else:
                    ci = ["", ""]#[1., 1., 1.]

                rows+=ci
            cols.append(rows)

        columns_array = np.array([item for sublist in [[col]*2 for col in self.frame.columns] for item in sublist] + ['r', 'CI'] * self.ncols).reshape(2, self.ncols*2)
        columns = pd.MultiIndex.from_tuples(list(zip(*columns_array)))#, names=["array", "value"])

        return pd.DataFrame(cols, index=self.columns, columns=columns)

    def _ci(self):
        
        cols = []
        for col in range(0, self.ncols):
            rows = []
            for row in range(0, self.ncols):

                # calculate r and ci
                if row != col:
                    r_value = (np.corrcoef(self.frame[self.columns[row]], self.frame[self.columns[col]])[0,1])
                        
                    z_l = 0.5 * (np.log(1+r_value) - np.log(1-r_value)) - self.z / (self.frame[[self.columns[row], self.columns[col]]].dropna(axis=0).shape[0]-3)
                    z_u = 0.5 * (np.log(1+r_value) - np.log(1-r_value)) + self.z / (self.frame[[self.columns[row], self.columns[col]]].dropna(axis=0).shape[0]-3)
                    rho_l = (np.exp(2 * z_l) - 1) / (np.exp(2 * z_l) + 1)
                    rho_u = (np.exp(2 * z_u) - 1) / (np.exp(2 * z_u) + 1)

                    ci = (rho_l, rho_u)
                else:
                    ci = (1.,1.)

                rows.append(ci)
            cols.append(rows)

        return pd.DataFrame(cols, index=self.columns, columns=self.columns)

    
    def _r(self):
        return pd.DataFrame([[(np.corrcoef(self.frame[col], self.frame[row])[0,1]) for col in self.columns] for row in self.columns], index=self.columns, columns=self.columns)



import bluebelt.styles.defaults as defaults

# main plot | pyplot.plot
plot = {
    #'color': defaults.black,
    'linestyle': 'solid',
    'linewidth': 3,
    'zorder': 90,
}

multiindex_label_line = {
    'color': defaults.black,
}
# plot title | pyplot.set_title
title = {
    'loc': 'left',
}